
from greenheart.simulation.technologies.hydrogen.h2_storage.pressure_vessel.tankinator import Tank, TypeITank, TypeIIITank, TypeIVTank
from greenheart.simulation.technologies.hydrogen.h2_storage.pressure_vessel import von_mises
import pytest
import numpy as np

# test that we the results we got when the code was recieved
class TestTankinator():

    def test_tank_type(self):
        tank1= TypeITank('6061_T6_Aluminum')
        tank3= TypeIIITank()
        tank4= TypeIVTank()

        assert tank1.tank_type == 1
        assert tank3.tank_type == 3
        assert tank4.tank_type == 4

    # def test_tank_type_exc(self):
    #     with pytest.raises(TypeError):
    #         TypeITank(2)

    def test_hemicylindrical_static(self):
        """ a random sphere, and a random cylinder """

        # L = 2R; should reduce to a sphere!
        radius= 1.5
        length= 2*radius
        thickness= 0.19
        volume_exact= 4./3.*np.pi*radius**3

        assert Tank.compute_hemicylinder_outer_length(length, thickness) == pytest.approx(length + 2.0*thickness)
        assert Tank.compute_hemicylinder_outer_radius(radius, thickness) == pytest.approx(radius + thickness)
        assert Tank.compute_hemicylinder_volume(radius, length) == pytest.approx(volume_exact)

        # a lil bit between sphere halves
        radius= 2.1
        length= 6.4
        thickness= 0.02
        volume_exact= 4./3.*np.pi*radius**3 + (length - 2*radius)*(np.pi*radius**2)

        assert Tank.compute_hemicylinder_outer_length(length, thickness) == pytest.approx(length + 2.0*thickness)
        assert Tank.compute_hemicylinder_outer_radius(radius, thickness) == pytest.approx(radius + thickness)
        assert Tank.compute_hemicylinder_volume(radius, length) == pytest.approx(volume_exact)

    def test_tankI_geometric(self):
        """ make sure geometric calls work correctly """

        # L = 2R; should reduce to a sphere!
        radius= 1.5
        length= 2*radius
        thickness= 0.19
        volume_exact= 4./3.*np.pi*radius**3

        # create tank, set dimensions, check dimensioning
        tank= TypeITank("6061_T6_Aluminum")
        tank.set_length_radius(length, radius)
        tank.thickness = thickness # manual override
        tank.volume_inner = Tank.compute_hemicylinder_volume(tank.get_radius_inner(),
                                                             tank.get_length_inner())

        assert tank.get_length_outer() == pytest.approx(length + 2.0*thickness)
        assert tank.get_radius_outer() == pytest.approx(radius + thickness)
        assert tank.get_volume_inner() == pytest.approx(volume_exact)

        volume_outer_exact= 4./3.*np.pi*(radius + thickness)**3
        dvolume_exact= volume_outer_exact - volume_exact
        rho_ref= 0.002663
        cost_ref= 4.45
        mass_exact= dvolume_exact*rho_ref # mass of spherical vessel
        cost_exact= mass_exact*cost_ref

        assert tank.get_volume_outer() == pytest.approx(volume_outer_exact)
        assert tank.get_volume_metal() == pytest.approx(dvolume_exact)
        assert tank.get_mass_metal() == pytest.approx(mass_exact)
        assert tank.get_cost_metal() == pytest.approx(cost_exact)

        assert tank.get_gravimetric_tank_efficiency() == pytest.approx((volume_exact/1e3)/mass_exact)
    
    def test_tankI_set_functions(self):
        """ make sure that the inverse geometry spec works """

        radius= 4.3
        length= 14.9
        volume_exact= 4./3.*np.pi*radius**3 + (length - 2*radius)*(np.pi*radius**2)

        tank= Tank(1, "316SS")
        tank.set_length_radius(length, radius)
        assert tank.get_volume_inner() == pytest.approx(volume_exact)
        
        tank.length_inner= tank.radius_inner= None # reset
        tank.set_length_volume(length, volume_exact)
        assert tank.get_radius_inner() == pytest.approx(radius)

        tank.length_inner= tank.radius_inner= None # reset
        tank.set_radius_volume(radius, volume_exact)
        assert tank.get_length_inner() == pytest.approx(length)

    def test_tankinator_typeI_comp(self):
        """ compare to the tankinator case """

        T_op= -50 # degC
        p_op= 170 # bar
        Ltank= 1000 # cm
        Vtank= 2994542 # ccm

        # reference values from the excel sheet default values
        R_ref= 31.2
        Sy_ref= 2953.475284
        Su_ref= 3327.58062
        density_ref= 0.002663
        costrate_ref= 4.45
        yield_thickness_ref= 2.69
        ultimate_thickness_ref= 3.586389441300914
        disp_vol_tw_ref= 7.462e5
        mass_tw_ref= 1987.08
        cost_tw_ref= 8842.51
        grav_eff_tw_ref= 1.51
        vmS1_0_ref= 1568.544508
        vmS2_0_ref= 699.2722538
        vmS3_0_ref= -170.
        vmSproof_0_ref= 2258.435564
        vmSburst_0_ref= 3387.653346
        WTAF_0_ref= 1.018052974
        thickness_1_ref= ultimate_thickness_ref*WTAF_0_ref
        vmS1_1_ref= 1542.397758
        vmS2_1_ref= 686.1988791
        vmS3_1_ref= -170.
        vmSproof_1_ref= 2224.46994
        vmSburst_1_ref= 3336.70491
        WTAF_1_ref= 1.002742019
        thickness_2_ref= thickness_1_ref*WTAF_1_ref
        thickness_f_ref= 3.6626944898294997
        vmS1_f_ref= 1537.826857
        vmS2_f_ref= 683.9134286
        vmS3_f_ref= -170.
        vmSproof_f_ref= 2218.532165
        vmSburst_f_ref= 3327.798248
        WTAF_f_ref= 1.000065401
        mass_f_ref= 2031.9
        cost_f_ref= 9041.80

        # set up w/ lookup shear approximation
        tank= TypeITank("6061_T6_Aluminum", shear_approx= "lookup")
        tank.set_operating_temperature(T_op)
        tank.set_operating_pressure(p_op)
        tank.set_length_volume(Ltank, Vtank)

        # check agains reference values
        assert tank.get_radius_inner() == pytest.approx(R_ref)
        assert tank.material.ultimate_shear_fun(T_op) == pytest.approx(Su_ref)
        assert tank.material.yield_shear_fun(T_op) == pytest.approx(Sy_ref)
        assert tank.material.density == pytest.approx(density_ref)
        assert tank.material.cost_rate == pytest.approx(costrate_ref)

        # check the thinwall calculations
        assert tank.get_yield_thickness() == pytest.approx(yield_thickness_ref, abs= 0.01)
        assert tank.get_ultimate_thickness() == pytest.approx(ultimate_thickness_ref, abs= 0.001)
        assert tank.get_thickness_thinwall() == pytest.approx(ultimate_thickness_ref, abs= 0.001)

        # check the implied geometry if we set the thickness to the thinwall
        tank.set_thickness_thinwall()
        assert tank.get_volume_metal() == pytest.approx(disp_vol_tw_ref, rel= 0.001)
        assert tank.get_mass_metal() == pytest.approx(mass_tw_ref, rel= 0.001)
        assert tank.get_cost_metal() == pytest.approx(cost_tw_ref)
        assert tank.get_gravimetric_tank_efficiency() == pytest.approx(grav_eff_tw_ref, abs= 0.01)

        # check von Mises analysis variables
        assert von_mises.S1(p_op, R_ref + ultimate_thickness_ref, R_ref) == pytest.approx(vmS1_0_ref)
        assert von_mises.S2(p_op, R_ref + ultimate_thickness_ref, R_ref) == pytest.approx(vmS2_0_ref)
        assert von_mises.S3(p_op, R_ref + ultimate_thickness_ref, R_ref) == pytest.approx(vmS3_0_ref)
        vmSproof, vmSburst= von_mises.getPeakStresses(p_op, R_ref + ultimate_thickness_ref, R_ref)
        assert vmSproof == pytest.approx(vmSproof_0_ref)
        assert vmSburst == pytest.approx(vmSburst_0_ref)
        assert not Tank.check_thinwall(R_ref, ultimate_thickness_ref)
        assert von_mises.wallThicknessAdjustmentFactor(p_op, R_ref + ultimate_thickness_ref, R_ref,
                                                       Sy_ref, Su_ref) == pytest.approx(WTAF_0_ref)
        
        # check cycle iterations, through two
        WTAF_0, thickness_1= von_mises.iterate_thickness(p_op, R_ref, ultimate_thickness_ref,
                                                         Sy_ref, Su_ref)
        assert WTAF_0 == pytest.approx(WTAF_0_ref)
        assert thickness_1 == pytest.approx(thickness_1_ref)
        
        WTAF_1, thickness_2= von_mises.iterate_thickness(p_op, R_ref, thickness_1,
                                                         Sy_ref, Su_ref)
        assert WTAF_1 == pytest.approx(WTAF_1_ref)
        assert thickness_2 == pytest.approx(thickness_2_ref)
        
        # check final value: cycle three times (no tol) to match tankinator
        (thickness_cycle, WTAF_cycle, n_iter)= von_mises.cycle(p_op, R_ref, ultimate_thickness_ref,
                                                               Sy_ref, Su_ref,
                                                               max_iter= 3, WTAF_tol= 0)
        
        print(thickness_cycle, WTAF_cycle, n_iter) # DEBUG
        assert thickness_cycle == pytest.approx(thickness_f_ref)

        # make sure final calculations are correct
        tank.set_thickness_vonmises(p_op, T_op, max_cycle_iter= 3, adj_fac_tol= 0.0)
        assert tank.get_thickness() == pytest.approx(thickness_f_ref)
        assert tank.get_mass_metal() == pytest.approx(mass_f_ref, abs= 0.1)
        assert tank.get_cost_metal() == pytest.approx(cost_f_ref, abs= 0.01)    

    def test_tankinator_typeIII_comp(self):
        """ compare to the tankinator case """

        T_op= 20. # degC
        p_op= 250. # bar
        Rtank= 16.0
        Ltank= 1219. # cm
        Vtank= 971799. # ccm

        # reference values from the excel sheet default values, best estimate
        R_ref= 16.0
        thickness_liner_ref= 0.61
        thickness_ideal_jacket_ref= 0.602759006
        Nlayer_jacket_ref= 7
        thickness_jacket_ref= 0.64008
        length_liner_ref= 1220.218176
        radius_liner_ref= 16.60908798
        V_outer_liner_ref= 1047.900361*1000
        V_liner_ref= 76101.03372
        m_liner_ref= 202.66
        cost_liner_ref= 901.82
        length_outer_ref= 1221.498336
        radius_outer_ref= 17.24916798
        V_outer_ref= 1131.022248*1000
        V_jacket_ref= 83121.88687
        m_jacket_ref= 133.91
        cost_jacket_ref= 4104.32
        m_tank_ref= 336.57
        cost_tank_ref= 5006.15
        gravimetric_tank_efficiency_ref= 2.89

        # set up w/ lookup shear approximation
        tank= TypeIIITank()
        tank.set_operating_temperature(T_op)
        tank.set_operating_pressure(p_op)
        tank.set_length_radius(Ltank, Rtank)

        tank.set_thicknesses_thinwall()

        # check against reference values
        assert tank.get_radius_inner() == pytest.approx(R_ref)
        assert tank.get_volume_inner() == pytest.approx(Vtank)
        assert tank.thickness_liner == pytest.approx(thickness_liner_ref, abs= 0.01)
        assert tank.thickness_ideal_jacket == pytest.approx(thickness_ideal_jacket_ref)
        assert tank.Nlayer_jacket == pytest.approx(Nlayer_jacket_ref)
        assert tank.thickness_jacket == pytest.approx(thickness_jacket_ref)

        assert tank.get_length_liner() == pytest.approx(length_liner_ref)
        assert tank.get_radius_liner() == pytest.approx(radius_liner_ref)
        assert tank.get_volume_outer_liner() == pytest.approx(V_outer_liner_ref)
        assert tank.get_volume_liner() == pytest.approx(V_liner_ref)

        assert tank.get_length_outer() == pytest.approx(length_outer_ref)
        assert tank.get_radius_outer() == pytest.approx(radius_outer_ref)
        assert tank.get_volume_outer() == pytest.approx(V_outer_ref)
        assert tank.get_volume_jacket() == pytest.approx(V_jacket_ref)

        assert tank.get_mass_liner() == pytest.approx(m_liner_ref, abs= 0.01)
        assert tank.get_mass_jacket() == pytest.approx(m_jacket_ref, abs= 0.01)
        assert tank.get_cost_liner() == pytest.approx(cost_liner_ref, abs= 0.01)
        assert tank.get_cost_jacket() == pytest.approx(cost_jacket_ref, abs= 0.01)
        assert tank.get_mass_tank() == pytest.approx(m_tank_ref, abs= 0.01)
        assert tank.get_cost_tank() == pytest.approx(cost_tank_ref, abs= 0.01)
        assert tank.get_gravimetric_tank_efficiency() == pytest.approx(gravimetric_tank_efficiency_ref, abs= 0.01)

    def test_tankinator_typeIV_comp(self):
        """ compare to the tankinator case """

        T_op= 20. # degC
        p_op= 350. # bar
        Rtank= 50.
        Ltank= 1000. # cm
        Vtank= 7592182. # ccm

        # reference values from the excel sheet default values, best estimate
        R_ref= 50.0
        thickness_liner_ref= 0.4
        thickness_ideal_jacket_ref= 3.241375931
        Nlayer_jacket_ref= 36
        thickness_jacket_ref= 3.29184
        # length_liner_ref= 1220.218176
        radius_liner_ref= 50.4
        # V_outer_liner_ref= 1047.900361*1000
        # V_liner_ref= 76101.03372
        # m_liner_ref= 202.66
        # cost_liner_ref= 901.82
        # length_outer_ref= 1221.498336
        # radius_outer_ref= 17.24916798
        # V_outer_ref= 1131.022248*1000
        # V_jacket_ref= 83121.88687
        # m_jacket_ref= 133.91
        # cost_jacket_ref= 4104.32
        # m_tank_ref= 336.57
        # cost_tank_ref= 5006.15
        # gravimetric_tank_efficiency_ref= 2.89

        # set up w/ lookup shear approximation
        tank= TypeIVTank()
        tank.set_operating_temperature(T_op)
        tank.set_operating_pressure(p_op)
        tank.set_length_radius(Ltank, Rtank)

        tank.set_thicknesses_thinwall()

        # check against reference values
        assert tank.get_radius_inner() == pytest.approx(R_ref)
        assert tank.get_volume_inner() == pytest.approx(Vtank)
        assert tank.thickness_liner == pytest.approx(thickness_liner_ref, abs= 0.01)
        assert tank.thickness_ideal_jacket == pytest.approx(thickness_ideal_jacket_ref)
        assert tank.Nlayer_jacket == pytest.approx(Nlayer_jacket_ref)
        assert tank.thickness_jacket == pytest.approx(thickness_jacket_ref)

        # assert tank.get_length_liner() == pytest.approx(length_liner_ref)
        assert tank.get_radius_liner() == pytest.approx(radius_liner_ref)
        # assert tank.get_volume_outer_liner() == pytest.approx(V_outer_liner_ref)
        # assert tank.get_volume_liner() == pytest.approx(V_liner_ref)

        # assert tank.get_length_outer() == pytest.approx(length_outer_ref)
        # assert tank.get_radius_outer() == pytest.approx(radius_outer_ref)
        # assert tank.get_volume_outer() == pytest.approx(V_outer_ref)
        # assert tank.get_volume_jacket() == pytest.approx(V_jacket_ref)

        # assert tank.get_mass_liner() == pytest.approx(m_liner_ref, abs= 0.01)
        # assert tank.get_mass_jacket() == pytest.approx(m_jacket_ref, abs= 0.01)
        # assert tank.get_cost_liner() == pytest.approx(cost_liner_ref, abs= 0.01)
        # assert tank.get_cost_jacket() == pytest.approx(cost_jacket_ref, abs= 0.01)
        # assert tank.get_mass_tank() == pytest.approx(m_tank_ref, abs= 0.01)
        # assert tank.get_cost_tank() == pytest.approx(cost_tank_ref, abs= 0.01)
        # assert tank.get_gravimetric_tank_efficiency() == pytest.approx(gravimetric_tank_efficiency_ref, abs= 0.01)

if __name__ == "__main__":
    test_set = test_tankinator()
    #test_set = TestTankinator()



import hopp.to_organize.H2_Analysis.H2AModel as H2AModel
import numpy as np
import pandas as pd


def run_h2a(electrical_generation_timeseries, kw_continuous, electrolyzer_size,
                                  hybrid_plant, reopt_results, scenario,
            combined_hybrid_curtailment_hopp, lcoe, force_electrolyzer_cost,
            forced_electrolyzer_cost_kw, total_system_electrical_usage=55.5):

    avg_generation = np.mean(electrical_generation_timeseries)  # Avg Generation

    if avg_generation < kw_continuous:
        cap_factor = avg_generation / kw_continuous
    else:
        cap_factor = 0.97

    max_hourly_h2_production = (electrolyzer_size * 1000) * cap_factor / total_system_electrical_usage
    hydrogen_hourly_production = cap_factor * np.divide(electrical_generation_timeseries,
                                           total_system_electrical_usage)  # hourly hydrogen production (kg)
    hydrogen_hourly_production[hydrogen_hourly_production > max_hourly_h2_production] = max_hourly_h2_production

    # Get Daily Hydrogen Production - Add Every 24 hours
    i = 0
    daily_H2_production = []
    while i < 8760:
        x = sum(hydrogen_hourly_production[i:i + 25])
        daily_H2_production.append(x)
        i = i + 25

    avg_daily_H2_production = np.mean(daily_H2_production)  # kgH2/day
    hydrogen_annual_output = sum(hydrogen_hourly_production)  # kgH2/year
    elec_remainder_after_h2 = combined_hybrid_curtailment_hopp

    # Run H2A model
    H2A_Results = H2AModel.H2AModel(cap_factor, avg_daily_H2_production, hydrogen_annual_output, force_system_size=True,
                                   forced_system_size=electrolyzer_size, force_electrolyzer_cost=True,
                                    forced_electrolyzer_cost_kw=forced_electrolyzer_cost_kw)

    print("Peak Daily Production Rate for H2 Electrolyzer {}".format(H2A_Results['peak_daily_production_rate']))
    h2a_costs = H2A_Results['Total Hydrogen Cost ($/kgH2)']
    determined_electrolyzer_size = H2A_Results['electrolyzer_size']
    determined_electrolyzer_plant_total_size = H2A_Results['total_plant_size']
    h2_scaled_total_installed_cost = H2A_Results['scaled_total_installed_cost']
    h2_scaled_total_installed_cost_kw = H2A_Results['scaled_total_installed_cost_kw']

    feedstock_cost_h2_levelized_hopp = lcoe * total_system_electrical_usage / 100  # $/kg
    # Hybrid Plant - levelized H2 Cost - HOPP
    feedstock_cost_h2_via_net_cap_cost_lifetime_h2_hopp = hybrid_plant.grid.financial_model.Outputs.adjusted_installed_cost / \
                                                          (hydrogen_annual_output * scenario['Useful Life'])  # $/kgH2
    # Total Hydrogen Cost ($/kgH2)
    total_unit_cost_of_hydrogen = h2a_costs + feedstock_cost_h2_levelized_hopp

    # feedstock_cost_h2_via_net_cap_cost_lifetime_h2_hopp = hybrid_plant.grid.financial_model.Outputs.adjusted_installed_cost /\
    #                                                     ((kw_continuous/55.5)*(8760 * useful_life))
    feedstock_cost_h2_via_net_cap_cost_lifetime_h2_reopt = reopt_results['outputs']['Scenario']['Site'] \
                                                               ['Financial']['net_capital_costs'] / (
                                                                       (kw_continuous / total_system_electrical_usage) * (8760 * scenario['Useful Life']))
    H2_Results = {'hydrogen_annual_output':
                        hydrogen_annual_output,
                    'feedstock_cost_h2_levelized_hopp':
                       feedstock_cost_h2_levelized_hopp,
                   'feedstock_cost_h2_via_net_cap_cost_lifetime_h2_hopp':
                       feedstock_cost_h2_via_net_cap_cost_lifetime_h2_hopp,
                   'feedstock_cost_h2_via_net_cap_cost_lifetime_h2_reopt':
                       feedstock_cost_h2_via_net_cap_cost_lifetime_h2_reopt,
                   'total_unit_cost_of_hydrogen':
                       total_unit_cost_of_hydrogen,
                   'cap_factor':
                       cap_factor
                   }

    # resultsDF = pd.DataFrame(result_dict, index=[0])

    return H2_Results, H2A_Results


## Low-Temperature PEM Electrolyzer Model
"""
Python model of H2 PEM low-temp electrolyzer.

Quick Hydrogen Physics:

1 kg H2 <-> 11.1 N-m3 <-> 33.3 kWh (LHV) <-> 39.4 kWh (HHV)

High mass energy density (1 kg H2= 3,77 l gasoline)
Low volumetric density (1 Nm³ H2= 0,34 l gasoline

Hydrogen production from water electrolysis (~5 kWh/Nm³ H2)

Power:1 MW electrolyser <-> 200 Nm³/h  H2 <-> ±18 kg/h H2
Energy:+/-55 kWh of electricity --> 1 kg H2 <-> 11.1 Nm³ <-> ±10 liters
demineralized water

Power production from a hydrogen PEM fuel cell from hydrogen (+/-50%
efficiency):
Energy: 1 kg H2 --> 16 kWh
"""
import math
import numpy as np
import sys
import pandas as pd
from matplotlib import pyplot as plt
import scipy
import rainflow
from scipy import interpolate

np.set_printoptions(threshold=sys.maxsize)

# def calc_current(P_T,p1,p2,p3,p4,p5,p6): #calculates i-v curve coefficients given the stack power and stack temp
#     pwr,tempc=P_T
#     i_stack=p1*(pwr**2) + p2*(tempc**2)+ (p3*pwr*tempc) +  (p4*pwr) + (p5*tempc) + (p6)
#     return i_stack 
def calc_current(P_T,p1,p2,p3,p4,p5,p6): #calculates i-v curve coefficients given the stack power and stack temp
    pwr,tempc=P_T
    # i_stack=p1*(pwr**2) + p2*(tempc**2)+ (p3*pwr*tempc) +  (p4*pwr) + (p5*tempc) + (p6)
    i_stack=p1*(pwr**3) + p2*(pwr**2) +  (p3*pwr) + (p4*pwr**(1/2)) + p5
    return i_stack 

class PEM_H2_Clusters:
    """
    Create an instance of a low-temperature PEM Electrolyzer System. Each
    stack in the electrolyzer system in this model is rated at 1 MW_DC.

    Parameters
    _____________
    np_array P_input_external_kW
        1-D array of time-series external power supply

    string voltage_type
        Nature of voltage supplied to electrolyzer from the external power
        supply ['variable' or 'constant]

    float power_supply_rating_MW
        Rated power of external power supply

    Returns
    _____________

    """
    #Remove: estimate_lifetime_capacity_factor
    #Remove: make_lifetime_performance_df_all_opt [x]
    def __init__(self, cluster_size_mw, plant_life, eol_eff_percent_loss=10,uptime_hours_until_eol = 77600, include_degradation_penalty=True,turndown_ratio=0.1,dt=3600):
        #self.input_dict = input_dict
        # print('RUNNING CLUSTERS PEM')
        self.set_max_h2_limit = False # TODO: add as input
        self.plant_life_years = plant_life
        
        self.include_deg_penalty = include_degradation_penalty
        self.use_onoff_deg = True
        self.use_uptime_deg= True
        self.use_fatigue_deg = True

        self.output_dict = {}
        self.dt=dt
        self.max_stacks = cluster_size_mw
        
        # self.stack_input_current_lower_bound = 400 #[A] any current below this amount (10% rated) will saturate the H2 production to zero, used to be 500 (12.5% of rated)
        self.stack_rating_kW = 1000  # 1 MW
        self.cell_active_area = 1920#1250 #[cm^2]
        self.N_cells = 130
        self.membrane_thickness=0.018 #cm
        self.cell_max_current_density = 2 #[A/cm^2]
        self.max_cell_current=self.cell_max_current_density*self.cell_active_area #PEM electrolyzers have a max current density of approx 2 A/cm^2 so max current is 2*cell_area
        # self.stack_input_current_lower_bound = 0.1*self.max_cell_current
        self.stack_input_current_lower_bound = turndown_ratio*self.max_cell_current
        
        self.turndown_ratio = turndown_ratio

        # Constants:
        self.moles_per_g_h2 = 0.49606 #[1/weight_h2]
        self.V_TN = 1.48  # Thermo-neutral Voltage (Volts) in standard conditions
        self.F = 96485.34  # Faraday's Constant (C/mol) or [As/mol]
        self.R = 8.314  # Ideal Gas Constant (J/mol/K)
        self.eta_h2_hhv=39.41

        #Additional Constants
        self.T_C = 80 #stack temperature in [C]
        self.mmHg_2_Pa = 133.322 #convert between mmHg to Pa
        self.patmo = 101325 #atmospheric pressure [Pa]
        self.mmHg_2_atm = self.mmHg_2_Pa/self.patmo #convert from mmHg to atm

        #Default degradation values
        
        self.onoff_deg_rate=1.47821515e-04 #[V/off-cycle]
        self.rate_fatigue = 3.33330244e-07 #multiply by rf_track
        
        self.curve_coeff=self.iv_curve() #this initializes the I-V curve to calculate current
        

        self.make_BOL_efficiency_curve()
        # if user_defined_EOL_percent_eff_loss:
        self.eol_eff_drop = eol_eff_percent_loss/100
        self.d_eol=self.find_eol_voltage_val(eol_eff_percent_loss)
        self.find_eol_voltage_curve(eol_eff_percent_loss)
        self.steady_deg_rate=self.reset_uptime_degradation_rate(uptime_hours_until_eol)

        
    def run(self,input_external_power_kw):
        startup_time=600 #[sec]
        startup_ratio = 1-(startup_time/self.dt)
        #saturate input power at rated power
        input_power_kw = self.external_power_supply(input_external_power_kw)
        #determine whether stack is on or off
        #0: off 1: on
        self.cluster_status = self.system_design(input_power_kw,self.max_stacks)
        #calculate number of on/off cycles
        #no delay at beginning of sim
        cluster_cycling = [0] + list(np.diff(self.cluster_status)) 
        cluster_cycling = np.array(cluster_cycling)
        
        #how much to reduce h2 by based on cycling status
        h2_multiplier = np.where(cluster_cycling > 0, startup_ratio, 1)
        #number of "stacks" on at a single time
        self.n_stacks_op = self.max_stacks*self.cluster_status
        #n_stacks_op is now either number of pem per cluster or 0 if cluster is off!

        #split power evenly amongst all stacks
        power_per_stack = np.where(self.n_stacks_op>0,input_power_kw/self.n_stacks_op,0)
        #calculate current from power
        stack_current =calc_current((power_per_stack,self.T_C), *self.curve_coeff)

        if self.include_deg_penalty:
            V_init=self.cell_design(self.T_C,stack_current)
            _,deg_signal=self.full_degradation(V_init)
            
            stack_current=self.find_equivalent_input_power_4_deg(power_per_stack,V_init,deg_signal) #fixed
            V_cell_equiv = self.cell_design(self.T_C,stack_current) #mabye this isn't necessary
            V_cell = V_cell_equiv + deg_signal
        else:
            V_init=self.cell_design(self.T_C,stack_current)
            _,deg_signal=self.full_degradation(V_init)
            # lifetime_performance_df =self.make_lifetime_performance_df_all_opt(deg_signal,V_init,power_per_stack)
            V_cell=self.cell_design(self.T_C,stack_current) #+self.total_Vdeg_per_hr_sys
            
        
        time_until_replacement,stack_life = self.calc_stack_replacement_info(deg_signal)
        annual_performance = self.make_yearly_performance_dict(power_per_stack,deg_signal,V_init,I_op=[],grid_connected=False)
        # self.make_yearly_performance_dict(power_per_stack,lifetime_performance_df['Time until replacement [hours]'],deg_signal,V_init) #TESTING
        stack_power_consumed = (stack_current * V_cell * self.N_cells)/1000
        system_power_consumed = self.n_stacks_op*stack_power_consumed
        
        h2_kg_hr_system_init = self.h2_production_rate(stack_current,self.n_stacks_op)
        # h20_gal_used_system=self.water_supply(h2_kg_hr_system_init)
        p_consumed_max,rated_h2_hr = self.rated_h2_prod()
        #scales h2 production to account for start-up time if going from off->on
        h2_kg_hr_system = h2_kg_hr_system_init * h2_multiplier 

        h20_gal_used_system=self.water_supply(h2_kg_hr_system)

        pem_cf = np.sum(h2_kg_hr_system)/(rated_h2_hr*len(input_power_kw)*self.max_stacks)
        efficiency = self.system_efficiency(input_power_kw,stack_current) #Efficiency as %-HHV
        
        h2_results={}
        h2_results_aggregates={}
        h2_results['Input Power [kWh]'] = input_external_power_kw
        h2_results['hydrogen production no start-up time']=h2_kg_hr_system_init
        h2_results['hydrogen_hourly_production']=h2_kg_hr_system
        h2_results['water_hourly_usage_kg'] =h20_gal_used_system*3.79
        h2_results['electrolyzer_total_efficiency_perc'] = efficiency
        h2_results['kwh_per_kgH2'] = input_power_kw / h2_kg_hr_system
        h2_results['Power Consumed [kWh]'] = system_power_consumed
        h2_results_aggregates['Warm-Up Losses on H2 Production'] = np.sum(h2_kg_hr_system_init) - np.sum(h2_kg_hr_system)
        
        
        h2_results_aggregates['Stack Life [hours]'] = stack_life
        h2_results_aggregates['Time until replacement [hours]'] = time_until_replacement
        h2_results_aggregates['Stack Rated Power Consumed [kWh]'] = p_consumed_max
        h2_results_aggregates['Stack Rated H2 Production [kg/hr]'] = rated_h2_hr
        h2_results_aggregates['Cluster Rated Power Consumed [kWh]'] = p_consumed_max*self.max_stacks
        h2_results_aggregates['Cluster Rated H2 Production [kg/hr]'] = rated_h2_hr*self.max_stacks
        h2_results_aggregates['gal H20 per kg H2'] = np.sum(h20_gal_used_system)/np.sum(h2_kg_hr_system)
        h2_results_aggregates['Stack Rated Efficiency [kWh/kg]'] = p_consumed_max/rated_h2_hr
        h2_results_aggregates['Cluster Rated H2 Production [kg/yr]'] = rated_h2_hr*len(input_power_kw)*self.max_stacks
        h2_results_aggregates['Operational Time / Simulation Time (ratio)'] = self.percent_of_sim_operating #added
        h2_results_aggregates['Fraction of Life used during sim'] = self.frac_of_life_used #added
        #TODO: add results for stack replacement stuff based on RATED voltage, not distribution

        # h2_results_aggregates['Number of Lifetime Cluster Replacements'] = nsr_life
        h2_results_aggregates['PEM Capacity Factor (simulation)'] = pem_cf
        
        h2_results_aggregates['Total H2 Production [kg]'] =np.sum(h2_kg_hr_system)
        h2_results_aggregates['Total Input Power [kWh]'] =np.sum(input_external_power_kw)
        h2_results_aggregates['Total kWh/kg'] =np.sum(input_external_power_kw)/np.sum(h2_kg_hr_system)
        h2_results_aggregates['Total Uptime [sec]'] = np.sum(self.cluster_status * self.dt)
        h2_results_aggregates['Total Off-Cycles'] = np.sum(self.off_cycle_cnt)
        h2_results_aggregates['Final Degradation [V]'] =self.cumulative_Vdeg_per_hr_sys[-1]
        # h2_results_aggregates['IV curve coeff'] = self.curve_coeff
        # h2_results_aggregates.update(lifetime_performance_df.to_dict()) 
        h2_results_aggregates['Performance By Year'] = annual_performance #double check if errors
        
        []
        return h2_results, h2_results_aggregates
        # return h2_results_aggregates
   

    def find_eol_voltage_curve(self,eol_eff_percent_loss):
        eol_eff_mult = (100+eol_eff_percent_loss)/100
        i_bol = self.output_dict['BOL Efficiency Curve Info']['Current'].values
        V_bol = self.output_dict['BOL Efficiency Curve Info']['Cell Voltage'].values
        h2_bol = self.output_dict['BOL Efficiency Curve Info']['H2 Produced'].values
        h2_eol = h2_bol/eol_eff_mult

        i_eol_no_faradaic_loss=(h2_eol*1000*2*self.F*self.moles_per_g_h2)/(1*self.N_cells*self.dt)
        n_f=self.faradaic_efficiency(i_eol_no_faradaic_loss)
        i_eol =  (h2_eol*1000*2*self.F*self.moles_per_g_h2)/(n_f*self.N_cells*self.dt)
        self.d_eol_curve=((i_bol*V_bol)/i_eol) - V_bol #simple method
        
        

    def find_equivalent_input_power_4_deg(self,power_in_kW,V_init,V_deg):
        #TODO: rename function since it outputs current
        I_in = calc_current((power_in_kW,self.T_C), *self.curve_coeff)
        eff_mult = (V_init + V_deg)/V_init #(1 + eff drop)
        I_deg = I_in/eff_mult

        return I_deg



    def full_degradation(self,voltage_signal):
        #TODO: add reset if hits end of life degradation limit!
        voltage_signal = voltage_signal*self.cluster_status
        if self.use_uptime_deg:
            V_deg_uptime = self.calc_uptime_degradation(voltage_signal)
        else:
            V_deg_uptime=np.zeros(len(voltage_signal))
        if self.use_onoff_deg:
            V_deg_onoff = self.calc_onoff_degradation()
        else:
            V_deg_onoff = np.zeros(len(voltage_signal))
        
        V_signal = voltage_signal + np.cumsum(V_deg_uptime) + np.cumsum(V_deg_onoff)
        if self.use_fatigue_deg:
            V_fatigue=self.approx_fatigue_degradation(V_signal)
        else:
            V_fatigue=np.zeros(len(voltage_signal))
        deg_signal = np.cumsum(V_deg_uptime) + np.cumsum(V_deg_onoff) + V_fatigue

        
        self.cumulative_Vdeg_per_hr_sys=deg_signal
        voltage_final=voltage_signal + deg_signal
        
        self.output_dict['Cumulative Degradation Breakdown']=pd.DataFrame({'Uptime':np.cumsum(V_deg_uptime),'On/off':np.cumsum(V_deg_onoff),'Fatigue':V_fatigue})
        return voltage_final, deg_signal
   
    
    def calc_stack_replacement_info(self,deg_signal):
        """Stack life optimistic estimate based on rated efficiency"""
        #[V] degradation at end of simulation
        d_sim = deg_signal[-1] 
        
        #fraction of life that has been "spent" during simulation
        frac_of_life_used = d_sim/self.d_eol_curve[-1]
        #number of hours simulated
        sim_time_dt = len(deg_signal) 
        #number of hours operating
        operational_time_dt=np.sum(self.cluster_status) 
        
        #time between replacement [hrs] based on number of hours operating
        t_eod_operation_based = (1/frac_of_life_used)*operational_time_dt #[hrs]
        #time between replacement [hrs] based on simulation length
        t_eod_existance_based = (1/frac_of_life_used)*sim_time_dt
        #CF shouldn't be different
        #just report out one capacity factor
        self.frac_of_life_used = frac_of_life_used
        self.percent_of_sim_operating = operational_time_dt/sim_time_dt

        return t_eod_existance_based,t_eod_operation_based
    
    def make_yearly_performance_dict(self,power_in_kW,V_deg,V_cell,I_op,grid_connected):
        #NOTE: this is not the most accurate for cases where simulation length is not close to 8760
        #I_op only needed if grid connected, should be singular value
        refturb_schedule = np.zeros(self.plant_life_years)
        # refturb_period=int(np.floor(time_between_replacements/8760))
        # refturb_schedule[refturb_period:int(self.plant_life_years):refturb_period]=1
        
        sim_length = len(V_cell)
        
        death_threshold = self.d_eol_curve[-1]
        
        cluster_cycling = [0] + list(np.diff(self.cluster_status)) #no delay at beginning of sim
        cluster_cycling = np.array(cluster_cycling)
        startup_ratio = 1-(600/3600)#TODO: don't have this hard-coded
        h2_multiplier = np.where(cluster_cycling > 0, startup_ratio, 1)
        
        _,rated_h2_pr_stack_BOL=self.rated_h2_prod()
        rated_h2_pr_sim = rated_h2_pr_stack_BOL*self.max_stacks*sim_length

        kg_h2_pr_sim = np.zeros(int(self.plant_life_years))
        capfac_per_sim = np.zeros(int(self.plant_life_years))
        d_sim = np.zeros(int(self.plant_life_years))
        power_pr_yr_kWh = np.zeros(int(self.plant_life_years))
        Vdeg0 = 0
        
        for i in range(int(self.plant_life_years)): #assuming sim is close to a year
            V_deg_pr_sim = Vdeg0 + V_deg
            
            # it_died = any(V_deg_pr_sim>death_threshold)
            if np.max(V_deg_pr_sim)>death_threshold:
                #it died
                idx_dead = np.argwhere(V_deg_pr_sim>death_threshold)[0][0]
                V_deg_pr_sim = np.concatenate([V_deg_pr_sim[0:idx_dead],V_deg[idx_dead:sim_length]])
                # i_sim_dead = i
                refturb_schedule[i]=self.max_stacks
                
            if not grid_connected:
                stack_current = self.find_equivalent_input_power_4_deg(power_in_kW,V_cell,V_deg_pr_sim)
                h2_kg_hr_system_init = self.h2_production_rate(stack_current,self.n_stacks_op)
                # total_sim_input_power = self.max_stacks*np.sum(power_in_kW)
                power_pr_yr_kWh[i] = self.max_stacks*np.sum(power_in_kW)
            else:
                h2_kg_hr_system_init = self.h2_production_rate(I_op,self.n_stacks_op)
                h2_kg_hr_system_init = h2_kg_hr_system_init*np.ones(len(power_in_kW))
                annual_power_consumed_kWh = self.max_stacks*I_op*(V_cell + V_deg_pr_sim)*self.N_cells/1000
                # total_sim_input_power = np.sum(annual_power_consumed_kWh)
                power_pr_yr_kWh[i] = np.sum(annual_power_consumed_kWh)

            h2_kg_hr_system = h2_kg_hr_system_init*h2_multiplier
            kg_h2_pr_sim[i] = np.sum(h2_kg_hr_system)
            capfac_per_sim[i] = np.sum(h2_kg_hr_system)/rated_h2_pr_sim
            d_sim[i] = V_deg_pr_sim[sim_length-1]
            Vdeg0 = V_deg_pr_sim[sim_length-1]
        performance_by_year = {}
        year = np.arange(0,int(self.plant_life_years),1)
        
        performance_by_year['Capacity Factor [-]'] = dict(zip(year,capfac_per_sim))
        performance_by_year['Refurbishment Schedule [MW replaced/year]'] = dict(zip(year,refturb_schedule))
        performance_by_year['Annual H2 Production [kg/year]'] = dict(zip(year,kg_h2_pr_sim))
        performance_by_year['Annual Average Efficiency [kWh/kg]'] = dict(zip(year,power_pr_yr_kWh/kg_h2_pr_sim))
        performance_by_year['Annual Average Efficiency [%-HHV]'] = dict(zip(year,self.eta_h2_hhv/(power_pr_yr_kWh/kg_h2_pr_sim)))
        performance_by_year['Annual Energy Used [kWh/year]'] = dict(zip(year,power_pr_yr_kWh))

        return performance_by_year

        
    def reset_uptime_degradation_rate(self,uptime_hours_until_eol):
        #this re-calculates the uptime
        #degradation rate to align with 80,000 operational hrs at 0.97 CF 
        #of no fatigue and no on/off cycles to be end-of-life
        I_max = calc_current((self.stack_rating_kW,self.T_C),*self.curve_coeff)
        V_rated = self.cell_design(self.T_C,I_max)
        new_deg_rate=self.d_eol/(V_rated*uptime_hours_until_eol*3600)
        return new_deg_rate

    def calc_uptime_degradation(self,voltage_signal):
        #steady_deg_rate = 1.12775521e-09
        steady_deg_per_hr=self.dt*self.steady_deg_rate*voltage_signal*self.cluster_status
        cumulative_Vdeg=np.cumsum(steady_deg_per_hr)
        self.output_dict['Total Uptime [sec]'] = np.sum(self.cluster_status * self.dt)
        self.output_dict['Total Uptime Degradation [V]'] = cumulative_Vdeg[-1]

        return steady_deg_per_hr
        
    def calc_onoff_degradation(self):
        
        
        change_stack=np.diff(self.cluster_status)
        cycle_cnt = np.where(change_stack < 0, -1*change_stack, 0)
        cycle_cnt = np.array([0] + list(cycle_cnt))
        self.off_cycle_cnt = cycle_cnt
        stack_off_deg_per_hr= self.onoff_deg_rate*cycle_cnt
        self.output_dict['System Cycle Degradation [V]'] = np.cumsum(stack_off_deg_per_hr)[-1]
        self.output_dict['Off-Cycles'] = cycle_cnt
        return stack_off_deg_per_hr

    def approx_fatigue_degradation(self,voltage_signal,dt_fatigue_calc_hrs=168):
        #should not use voltage values when voltage_signal = 0
        #aka - should only be counted when electrolyzer is on
        
        t_calc=np.arange(0,len(voltage_signal)+dt_fatigue_calc_hrs ,dt_fatigue_calc_hrs ) 
        v_max=np.max(voltage_signal)
        v_min=np.min(voltage_signal)
        if v_max==v_min:
            rf_sum=0
            lifetime_fatigue_deg=0
            V_fatigue_ts=np.zeros(len(voltage_signal))

        else:

            rf_cycles = rainflow.count_cycles(voltage_signal, nbins=10)
            rf_sum = np.sum([pair[0] * pair[1] for pair in rf_cycles])
            lifetime_fatigue_deg=rf_sum*self.rate_fatigue
            self.output_dict['Approx Total Fatigue Degradation [V]'] = lifetime_fatigue_deg
            rf_track=0
            V_fatigue_ts=np.zeros(len(voltage_signal))
            for i in range(len(t_calc)-1):
                voltage_signal_temp = voltage_signal[np.nonzero(voltage_signal[t_calc[i]:t_calc[i+1]])]
                if np.size(voltage_signal_temp)==0:
                    rf_sum = 0
                else:
                    v_max=np.max(voltage_signal_temp)
                    v_min=np.min(voltage_signal_temp)

                    if v_max == v_min:
                        rf_sum=0
                    else:
                        
                        rf_cycles=rainflow.count_cycles(voltage_signal_temp, nbins=10)
                    # rf_cycles=rainflow.count_cycles(voltage_signal[t_calc[i]:t_calc[i+1]], nbins=10)
                        rf_sum=np.sum([pair[0] * pair[1] for pair in rf_cycles])
                rf_track+=rf_sum
                V_fatigue_ts[t_calc[i]:t_calc[i+1]]=rf_track*self.rate_fatigue
                #already cumulative!
            self.output_dict['Sim End RF Track'] = rf_track #TODO: remove
            self.output_dict['Total Actual Fatigue Degradation [V]'] = V_fatigue_ts[-1] #TODO: remove

        return V_fatigue_ts #already cumulative!

    def grid_connected_func(self,h2_kg_hr_system_required):
        """
        Calculate power and current required to meet a constant hydrogen demand
        """
        # df=self.output_dict['BOL Efficiency Curve Info'][['H2 Produced','Current','Power Sent [kWh]','Power Consumed [kWh]']]

        max_h2kg_single_stack=self.h2_production_rate(self.max_cell_current,1)
        # EOL_max_h2_stack=self.h2_production_rate(self.max_cell_current,1)
        min_n_stacks=np.ceil(h2_kg_hr_system_required/max_h2kg_single_stack)
        if min_n_stacks>self.max_stacks:
            print("ISSUE")
        h2_per_stack_min=h2_kg_hr_system_required/self.max_stacks #change var name
        
        I_reqd_BOL_noFaradaicLoss=(h2_per_stack_min*1000*2*self.F*self.moles_per_g_h2)/(1*self.N_cells*self.dt)
        n_f=self.faradaic_efficiency(I_reqd_BOL_noFaradaicLoss)
        I_reqd=(h2_per_stack_min*1000*2*self.F*self.moles_per_g_h2)/(n_f*self.N_cells*self.dt)
        V_reqd = self.cell_design(self.T_C,I_reqd)
        #TODO: only include deg if user-requested
        V_deg_per_hr=self.steady_deg_rate*V_reqd*self.dt
        V_steady_deg=np.arange(0,self.d_eol+V_deg_per_hr,V_deg_per_hr)
        P_reqd_per_hr_stack=I_reqd*(V_reqd + V_steady_deg)*self.N_cells/1000 #kW
        P_required_per_hr_system=self.max_stacks*P_reqd_per_hr_stack #kW

        output_system_power = P_required_per_hr_system[0:8760]
        stack_current_signal = I_reqd*np.ones(len(output_system_power))
        return output_system_power, stack_current_signal


    def find_eol_voltage_val(self,eol_rated_eff_drop_percent):
        #TODO: change this 
        rated_power_idx=self.output_dict['BOL Efficiency Curve Info'].index[self.output_dict['BOL Efficiency Curve Info']['Power Sent [kWh]']==self.stack_rating_kW].to_list()[0]
        rated_eff_df=self.output_dict['BOL Efficiency Curve Info'].iloc[rated_power_idx]
        i_rated=rated_eff_df['Current']
        h2_rated_kg=rated_eff_df['H2 Produced']
        vcell_rated=rated_eff_df['Cell Voltage']
        bol_eff_kWh_per_kg=rated_eff_df['Efficiency [kWh/kg]']
        eol_eff_kWh_per_kg=bol_eff_kWh_per_kg*(1+eol_rated_eff_drop_percent/100)
        eol_power_consumed_kWh = eol_eff_kWh_per_kg*h2_rated_kg
        v_tot_eol=eol_power_consumed_kWh*1000/(self.N_cells*i_rated)
        d_eol = v_tot_eol - vcell_rated
        return d_eol


    def system_efficiency(self,P_sys,I):
        #
        system_h2_prod_rate=self.h2_production_rate(I,self.n_stacks_op)
        eff_kWh_pr_kg = P_sys/system_h2_prod_rate #kWh/kg
        system_eff= self.eta_h2_hhv/eff_kWh_pr_kg #[%-HHV]
        
        return system_eff #[%]

    def make_BOL_efficiency_curve(self):
        #TODO: remove all other function calls to self.output_dict['BOL Efficiency Curve Info']
        #this should be done differntly
        # power_in_signal=np.arange(0.1,1.1,0.1)*self.stack_rating_kW
        power_in_signal=np.arange(self.turndown_ratio,1.1,0.1)*self.stack_rating_kW
        stack_I = calc_current((power_in_signal,self.T_C),*self.curve_coeff)
        stack_V = self.cell_design(self.T_C,stack_I)
        power_used_signal = (stack_I*stack_V*self.N_cells)/1000
        h2_stack_kg= self.h2_production_rate(stack_I ,1)
        kWh_per_kg=power_in_signal/h2_stack_kg
        power_error_BOL = power_in_signal-power_used_signal
        self.BOL_powerIn2error = interpolate.interp1d(power_in_signal,power_error_BOL)
        data=pd.DataFrame({'Power Sent [kWh]':power_in_signal,'Current':stack_I,'Cell Voltage':stack_V,
        'H2 Produced':h2_stack_kg,'Efficiency [kWh/kg]':kWh_per_kg,
        'Power Consumed [kWh]':power_used_signal,'IV Curve BOL Error [kWh]':power_error_BOL})
        self.output_dict['BOL Efficiency Curve Info']=data
        
        #return kWh_per_kg, power_in_signal



    def rated_h2_prod(self):
        i=self.output_dict['BOL Efficiency Curve Info'].index[self.output_dict['BOL Efficiency Curve Info']['Power Sent [kWh]']==self.stack_rating_kW]
        I_max=self.output_dict['BOL Efficiency Curve Info']['Current'].iloc[i].values[0]
        #I_max = calc_current((self.stack_rating_kW,self.T_C),*self.curve_coeff)
        V_max = self.cell_design(self.T_C,I_max)
        P_consumed_stack_kw = I_max*V_max*self.N_cells/1000
        max_h2_stack_kg= self.h2_production_rate(I_max,1)
        return P_consumed_stack_kw,max_h2_stack_kg

        
    def external_power_supply(self,input_external_power_kw):
        """
        External power source (grid or REG) which will need to be stepped
        down and converted to DC power for the electrolyzer.

        Please note, for a wind farm as the electrolyzer's power source,
        the model assumes variable power supplied to the stack at fixed
        voltage (fixed voltage, variable power and current)

        TODO: extend model to accept variable voltage, current, and power
        This will replicate direct DC-coupled PV system operating at MPP
        """
        power_converter_efficiency = 1.0 #this used to be 0.95 but feel free to change as you'd like
        # if self.input_dict['voltage_type'] == 'constant':
        power_curtailed_kw=np.where(input_external_power_kw > self.max_stacks * self.stack_rating_kW,\
        input_external_power_kw - self.max_stacks * self.stack_rating_kW,0)

        input_power_kw = \
            np.where(input_external_power_kw >
                        (self.max_stacks * self.stack_rating_kW),
                        (self.max_stacks * self.stack_rating_kW),
                        input_external_power_kw)

        self.output_dict['Curtailed Power [kWh]'] = power_curtailed_kw
        
        return input_power_kw
        # else:
        #     pass  # TODO: extend model to variable voltage and current source
    def iv_curve(self):
        """
        This is a new function that creates the I-V curve to calculate current based
        on input power and electrolyzer temperature

        current range is 0: max_cell_current+10 -> PEM have current density approx = 2 A/cm^2

        temperature range is 40 degC : rated_temp+5 -> temperatures for PEM are usually within 60-80degC

        calls cell_design() which calculates the cell voltage
        """
        # current_range = np.arange(0,self.max_cell_current+10,10) 
        current_range = np.arange(self.stack_input_current_lower_bound,self.max_cell_current+10,10) 
        temp_range = np.arange(40,self.T_C+5,5)
        idx = 0
        powers = np.zeros(len(current_range)*len(temp_range))
        currents = np.zeros(len(current_range)*len(temp_range))
        temps_C = np.zeros(len(current_range)*len(temp_range))
        for i in range(len(current_range)):
            
            for t in range(len(temp_range)):
                powers[idx] = current_range[i]*self.cell_design(temp_range[t],current_range[i])*self.N_cells*(1e-3) #stack power
                currents[idx] = current_range[i]
                temps_C[idx] = temp_range[t]
                idx = idx+1
        df=pd.DataFrame({'Power':powers,'Current':currents,'Temp':temps_C}) #added
        temp_oi_idx = df.index[df['Temp']==self.T_C]      #added  
        # curve_coeff, curve_cov = scipy.optimize.curve_fit(calc_current, (powers,temps_C), currents, p0=(1.0,1.0,1.0,1.0,1.0,1.0)) #updates IV curve coeff
        curve_coeff, curve_cov = scipy.optimize.curve_fit(calc_current, (df['Power'][temp_oi_idx].values,df['Temp'][temp_oi_idx].values), df['Current'][temp_oi_idx].values, p0=(1.0,1.0,1.0,1.0,1.0,1.0))
        return curve_coeff

    def system_design(self,input_power_kw,cluster_size_mw):
        """
        Calculate whether the cluster is on or off based on input power

        TODO: add 0.1 (default turndown ratio) as input
        """
        # cluster_min_power = 0.1*self.max_stacks
        # cluster_min_power = 0.1*cluster_size_mw
        cluster_min_power = self.turndown_ratio*cluster_size_mw
        cluster_status=np.where(input_power_kw<cluster_min_power,0,1)
       
        return cluster_status

    def cell_design(self, Stack_T, Stack_Current):
        """
        Basically: calc_cell_voltage
        """
        E_cell = self.calc_reversible_cell_voltage(Stack_T)
        V_act = self.calc_V_act(Stack_T,Stack_Current,self.cell_active_area)
        V_ohmic=self.calc_V_ohmic(Stack_T,Stack_Current,self.cell_active_area,self.membrane_thickness)
        V_cell = E_cell + V_act + V_ohmic
       
        return V_cell
    def calc_reversible_cell_voltage(self,Stack_T):
        """
        inputs::
            Stack_T [C]: operating temperature
        
        returns:: 
            E_cell [V/cell]: reversible overpotential

        Reference:
        
        """
        T_K=Stack_T+ 273.15  # convert Celsius to Kelvin
        #Reversible potential at 25degC - Nerst Equation
        E_rev0 = 1.229  #[V] 
        panode_atm=1 #[atm] total pressure at the anode
        pcathode_atm=1 #[atm] total pressure at the cathode
        patmo_atm=1 #atmospheric prestture
    
        #coefficient for Antoine formulas
        A = 8.07131
        B = 1730.63
        C = 233.426

        #vapor pressure of water in [mmHg] using Antoine formula
        #valid for T<283 Kelvin
        p_h2o_sat_mmHg = 10 ** (A - (B / (C + Stack_T)))  
        #convert mmHg to atm
        p_h20_sat_atm=p_h2o_sat_mmHg*self.mmHg_2_atm  
        #p_H2 = p_cat - p_h2O
        #p_O2 = p_an - p_h2O
        # p_h2O_sat_Pa = (0.61121* np.exp((18.678 - (Stack_T / 234.5)) * (Stack_T / (257.14 + Stack_T)))) * 1e3  # (Pa) #ARDEN-BUCK
        # p_h20_sat_atm=p_h2O_sat_Pa/self.patmo
                # Cell reversible voltage kind of explain in Equations (12)-(15) of below source
        # https://www.sciencedirect.com/science/article/pii/S0360319906000693
        # OR see equation (8) in the source below
        # https://www.sciencedirect.com/science/article/pii/S0360319917309278?via%3Dihub
        #h2 outlet pressure would be 5000 kPa and O2 outlet pressure could be 200 kPa
        #Nerst Equation
        E_cell=E_rev0 + ((self.R*T_K)/(2*self.F))*(np.log(((panode_atm-p_h20_sat_atm)/patmo_atm)*np.sqrt((pcathode_atm-p_h20_sat_atm)/patmo_atm))) 
        return E_cell

    def calc_V_act(self,Stack_T,I_stack,cell_active_area):
        """
        inputs::
            stack_T [C]: operating temperature
            I_stack [A]: stack current
            cell_active_area [cm^2]: electrode area
        
        returns:: 
            V_act [V/cell]: anode and cathode activation overpotential

        Reference:
        
        """
        T_K=Stack_T+ 273.15 
        #current density [A/cm^2]
        i = I_stack/cell_active_area
        # Anode charge transfer coefficient
        a_a = 2  
        # Cathode charge transfer coefficient
        a_c = 0.5  
        #anode exchange current density
        i_o_a = 2 * (10 ** (-7)) 
        #cathode exchange current density
        i_o_c = 2 * (10 ** (-3)) 
        V_anode = (((self.R * T_K) / (a_a * self.F)) * np.arcsinh(i / (2 * i_o_a)))
        V_cathode= (((self.R * T_K) / (a_c * self.F)) * np.arcsinh(i / (2 * i_o_c)))
        V_act = V_anode + V_cathode
        return V_act

    def calc_V_ohmic(self,Stack_T,I_stack,cell_active_area,delta_cm):
        T_K=Stack_T+ 273.15 
        #current density [A/cm^2]
        i = I_stack/cell_active_area
        lambda_water_content = ((-2.89556 + (0.016 * T_K)) + 1.625) / 0.1875
        # membrane proton conductivity [S/cm]
        sigma = ((0.005139 * lambda_water_content) - 0.00326) * np.exp(
            1268 * ((1 / 303) - (1 / T_K)))   
        #ionic resistance [ohms*cm^2]
        R_cell = (delta_cm / sigma) 
        # [ohms*cm^2] from Table 1 in  https://journals.utm.my/jurnalteknologi/article/view/5213/3557
        R_elec=3.5*(10 ** (-5))
        V_ohmic=(i *( R_cell + R_elec)) 
        return V_ohmic
    def dynamic_operation(self): #UNUSED Here but capability is included
        """
        Model the electrolyzer's realistic response/operation under variable RE

        TODO: add this capability to the model
        """
        # When electrolyzer is already at or near its optimal operation
        # temperature (~80degC)
        
        warm_startup_time_secs = 30
        cold_startup_time_secs = 5 * 60  # 5 minutes


    def faradaic_efficiency(self,stack_current): #ONLY EFFICIENCY CONSIDERED RIGHT NOW
        """`
        Text background from:
        [https://www.researchgate.net/publication/344260178_Faraday%27s_
        Efficiency_Modeling_of_a_Proton_Exchange_Membrane_Electrolyzer_
        Based_on_Experimental_Data]

        In electrolyzers, Faraday’s efficiency is a relevant parameter to
        assess the amount of hydrogen generated according to the input
        energy and energy efficiency. Faraday’s efficiency expresses the
        faradaic losses due to the gas crossover current. The thickness
        of the membrane and operating conditions (i.e., temperature, gas
        pressure) may affect the Faraday’s efficiency.

        Equation for n_F obtained from:
        https://www.sciencedirect.com/science/article/pii/S0360319917347237#bib27

        Parameters
        ______________
        float f_1
            Coefficient - value at operating temperature of 80degC (mA2/cm4)

        float f_2
            Coefficient - value at operating temp of 80 degC (unitless)

        np_array current_input_external_Amps
            1-D array of current supplied to electrolyzer stack from external
            power source


        Returns
        ______________

        float n_F
            Faradaic efficiency (unitless)

        """
        f_1 = 250  # Coefficient (mA2/cm4)
        f_2 = 0.996  # Coefficient (unitless)
        I_cell = stack_current * 1000

        # Faraday efficiency
        n_F = (((I_cell / self.cell_active_area) ** 2) /
               (f_1 + ((I_cell / self.cell_active_area) ** 2))) * f_2

        return n_F

    


    def h2_production_rate(self,stack_current,n_stacks_op):
        """
        H2 production rate calculated using Faraday's Law of Electrolysis
        (https://www.sciencedirect.com/science/article/pii/S0360319917347237#bib27)

        Parameters
        _____________

        np_array
            1-D array of current supplied to electrolyzer stack from external
            power source


        Returns
        _____________

        """
        # Single stack calculations:
        n_Tot = self.faradaic_efficiency(stack_current)
        h2_production_rate = n_Tot * ((self.N_cells *
                                       stack_current) /
                                      (2 * self.F))  # mol/s
        h2_production_rate_g_s = h2_production_rate / self.moles_per_g_h2
        h2_produced_kg_hr = h2_production_rate_g_s * (self.dt/1000 ) #3.6 #Fixed: no more manual scaling
        self.output_dict['stack_h2_produced_g_s']= h2_production_rate_g_s
        self.output_dict['stack_h2_produced_kg_hr'] = h2_produced_kg_hr

        # Total electrolyzer system calculations:
        h2_produced_kg_hr_system = n_stacks_op  * h2_produced_kg_hr
        # h2_produced_kg_hr_system = h2_produced_kg_hr
        self.output_dict['h2_produced_kg_hr_system'] = h2_produced_kg_hr_system

        return h2_produced_kg_hr_system


    def water_supply(self,h2_kg_hr):
        """
        Calculate water supply rate based system efficiency and H2 production
        rate
        TODO: Add water-to-hydrogen ratio as input, currently hard-coded to 10
        """
        # ratio of water_used:h2_kg_produced depends on power source
        # h20_kg:h2_kg with PV 22-126:1 or 18-25:1 without PV but considering water deminersalisation
        # stoichometrically its just 9:1 but ... theres inefficiencies in the water purification process
        
        water_used_kg_hr_system = h2_kg_hr * 10
        self.output_dict['water_used_kg_hr'] = water_used_kg_hr_system
        self.output_dict['water_used_kg_annual'] = np.sum(water_used_kg_hr_system)
        water_used_gal_hr_system = water_used_kg_hr_system/3.79
        return water_used_gal_hr_system 

    
    def run_grid_connected_workaround(self,power_input_signal,current_signal):
        #power input signal is total system input power
        #current signal is current per stack
        startup_time=600 #[sec]
        startup_ratio = 1-(startup_time/self.dt)
        self.cluster_status = self.system_design(power_input_signal,self.max_stacks)
        self.n_stacks_op = self.max_stacks*self.cluster_status
        cluster_cycling = [0] + list(np.diff(self.cluster_status)) #no delay at beginning of sim
        cluster_cycling = np.array(cluster_cycling)
        power_per_stack = np.where(self.n_stacks_op>0,power_input_signal/self.n_stacks_op,0)
        
        h2_multiplier = np.where(cluster_cycling > 0, startup_ratio, 1)
        self.n_stacks_op = self.max_stacks*self.cluster_status

        V_init=self.cell_design(self.T_C,current_signal)
        V_cell_deg,deg_signal=self.full_degradation(V_init)
        
        time_until_replacement,stack_life = self.calc_stack_replacement_info(deg_signal)
        annual_performance= self.make_yearly_performance_dict(power_per_stack,deg_signal,V_init,current_signal[0],grid_connected=True) #TESTING
        
        stack_power_consumed = (current_signal * V_cell_deg * self.N_cells)/1000
        system_power_consumed = self.n_stacks_op*stack_power_consumed
        
        h2_kg_hr_system_init = self.h2_production_rate(current_signal,self.n_stacks_op)
        p_consumed_max,rated_h2_hr = self.rated_h2_prod()
        h2_kg_hr_system = h2_kg_hr_system_init * h2_multiplier #scales h2 production to account
        #for start-up time if going from off->on
        h20_gal_used_system=self.water_supply(h2_kg_hr_system)

        pem_cf = np.sum(h2_kg_hr_system)/(rated_h2_hr*len(power_input_signal)*self.max_stacks)
        efficiency = self.system_efficiency(power_input_signal,current_signal)

        h2_results={}
        h2_results_aggregates={}
        h2_results['Input Power [kWh]'] = power_input_signal
        h2_results['hydrogen production no start-up time']=h2_kg_hr_system_init
        h2_results['hydrogen_hourly_production']=h2_kg_hr_system
        # h2_results['water_hourly_usage_gal'] =h20_gal_used_system
        h2_results['water_hourly_usage_kg'] =h20_gal_used_system*3.79
        h2_results['electrolyzer_total_efficiency_perc'] = efficiency
        h2_results['kwh_per_kgH2'] = power_input_signal / h2_kg_hr_system
        h2_results['Power Consumed [kWh]'] = system_power_consumed
        h2_results_aggregates['Warm-Up Losses on H2 Production'] = np.sum(h2_kg_hr_system_init) - np.sum(h2_kg_hr_system)

        h2_results_aggregates['Stack Rated Power Consumed [kWh]'] = p_consumed_max
        h2_results_aggregates['Stack Rated H2 Production [kg/hr]'] = rated_h2_hr
        h2_results_aggregates['Cluster Rated Power Consumed [kWh]'] = p_consumed_max*self.max_stacks
        h2_results_aggregates['Cluster Rated H2 Production [kg/hr]'] = rated_h2_hr*self.max_stacks
        h2_results_aggregates['Stack Rated Efficiency [kWh/kg]'] = p_consumed_max/rated_h2_hr
        h2_results_aggregates['Cluster Rated H2 Production [kg/yr]'] = rated_h2_hr*len(power_input_signal)*self.max_stacks
        h2_results_aggregates['gal H20 per kg H2'] = np.sum(h20_gal_used_system)/np.sum(h2_kg_hr_system)
        h2_results_aggregates['PEM Capacity Factor (simulation)'] = pem_cf
        
        h2_results_aggregates['Operational Time / Simulation Time (ratio)'] = self.percent_of_sim_operating #added
        h2_results_aggregates['Fraction of Life used during sim'] = self.frac_of_life_used #added

        h2_results_aggregates['Total H2 Production [kg]'] =np.sum(h2_kg_hr_system)
        h2_results_aggregates['Total Input Power [kWh]'] =np.sum(power_input_signal)
        h2_results_aggregates['Total kWh/kg'] =np.sum(power_input_signal)/np.sum(h2_kg_hr_system)
        h2_results_aggregates['Total Uptime [sec]'] = np.sum(self.cluster_status * self.dt)
        h2_results_aggregates['Total Off-Cycles'] = np.sum(self.off_cycle_cnt)
        h2_results_aggregates['Final Degradation [V]'] =self.cumulative_Vdeg_per_hr_sys[-1]
        
        h2_results_aggregates['Performance By Year'] = annual_performance #double check if errors
        # h2_results_aggregates['Stack Life Summary'] = self.stack_life_opt

        h2_results_aggregates['Stack Life [hours]'] = stack_life
        h2_results_aggregates['Time until replacement [hours]'] = time_until_replacement

      
        []
        return h2_results, h2_results_aggregates

if __name__ == "__main__":
    cluster_size_mw = 1 #MW
    plant_life = 30 #years
    electrolyzer_model_parameters = {
        "eol_eff_percent_loss":10, #percent - for calculating EOL voltage and steady deg rate
        "uptime_hours_until_eol": 77600, #for calculating steady deg rate
        "include_degradation_penalty": True,
        "turndown_ratio":0.1,
    }
    # Create PEM and initialize parameters
    pem = PEM_H2_Clusters(cluster_size_mw, plant_life,**electrolyzer_model_parameters)
    # ----- Run off-grid case -----
    # Make a mock input power signal
    hours_in_year = 8760
    power_signal_kW_ramp_up = np.arange(0,1000*cluster_size_mw,20)
    power_signal_kW_ramp_down = np.flip(power_signal_kW_ramp_up)
    power_signal_kW_ramp_updown = np.concatenate((power_signal_kW_ramp_down,power_signal_kW_ramp_up))
    n_repeats = int(np.ceil(hours_in_year/len(power_signal_kW_ramp_updown)))
    power_signal_kW = np.tile(power_signal_kW_ramp_updown,n_repeats)
    # Run the electrolyzer and get outputs
    h2_results, h2_results_aggregates = pem.run(power_signal_kW)
    
    # ----- Run grid-connected case -----
    h2_kg_hr_system_required = 12 #kg-H2/hr
    target_h2_per_year = h2_kg_hr_system_required*hours_in_year
    power_required_kW, stack_current_signal = pem.grid_connected_func(h2_kg_hr_system_required)
    h2_results_grid, h2_results_aggregates_grid = pem.run_grid_connected_workaround(power_required_kW,stack_current_signal)
    
    #Refurbishment Schedule (RS) [MW/year] and Year of Replacement (YOR)
    offgrid_RS = np.array(list(h2_results_aggregates['Performance By Year']['Refurbishment Schedule [MW replaced/year]'].values()))
    offgrid_YOR = np.argwhere(offgrid_RS>0)[:,0]
    grid_RS = np.array(list(h2_results_aggregates_grid['Performance By Year']['Refurbishment Schedule [MW replaced/year]'].values()))
    grid_YOR = np.argwhere(grid_RS>0)[:,0]
    # ----- Plot and compare results -----
    year_of_operation = list(h2_results_aggregates['Performance By Year']['Capacity Factor [-]'].keys())
    fig,ax = plt.subplots(nrows=3,ncols=1,sharex=True,figsize = [6.4,7.2])
    

    # Annual Hydrogen Produced (AH2) [metric tons H2/year]
    offgrid_AH2 = np.array(list(h2_results_aggregates['Performance By Year']['Annual H2 Production [kg/year]'].values()))/1000
    grid_AH2 = np.array(list(h2_results_aggregates_grid['Performance By Year']['Annual H2 Production [kg/year]'].values()))/1000

    ax[0].plot(year_of_operation,offgrid_AH2,color='b',ls='-',lw=2,label='off-grid')
    ax[0].plot(year_of_operation,grid_AH2,color='r',ls='--',lw=2,label='grid-connected')
    ax[0].vlines(offgrid_YOR,np.zeros(len(offgrid_YOR)),offgrid_AH2[offgrid_YOR],colors='skyblue',ls='-',lw=0.5)
    ax[0].vlines(grid_YOR,np.zeros(len(grid_YOR)),grid_AH2[grid_YOR],colors='tomato',ls='--',lw=0.5)
    ax[0].set_ylabel('Annual Hydrogen Produced\n[metric tons/year]')
    ax[0].legend(loc='upper right')
    ax[0].set_ylim((0.8*np.min([offgrid_AH2,grid_AH2]),1.2*np.max([offgrid_AH2,grid_AH2])))

    # Annual Energy Used (AEU) [MWh/year]
    offgrid_AEU = np.array(list(h2_results_aggregates['Performance By Year']['Annual Energy Used [kWh/year]'].values()))/1000
    grid_AEU = np.array(list(h2_results_aggregates_grid['Performance By Year']['Annual Energy Used [kWh/year]'].values()))/1000
    ax[1].plot(year_of_operation,offgrid_AEU,color='b',ls='-',lw=2,label='_off-grid')
    ax[1].plot(year_of_operation,grid_AEU,color='r',ls='--',lw=2,label='_grid-connected')
    ax[1].vlines(offgrid_YOR,np.zeros(len(offgrid_YOR)),offgrid_AEU[offgrid_YOR],colors='skyblue',ls='-',lw=0.5,label = 'off-grid stack replacement')
    ax[1].vlines(grid_YOR,np.zeros(len(grid_YOR)),grid_AEU[grid_YOR],colors='tomato',ls='--',lw=0.5,label = 'grid-connected stack replacement')
    ax[1].set_ylabel('Annual Energy Used\n[MWh/year]')
    ax[1].legend(loc='upper right')
    ax[1].set_ylim((0.8*np.min([offgrid_AEU,grid_AEU]),1.2*np.max([offgrid_AEU,grid_AEU])))
    
    # Annual Average Conversion Efficiency (AACE) [kWh/kg]
    offgrid_AACE = np.array(list(h2_results_aggregates['Performance By Year']['Annual Average Efficiency [kWh/kg]'].values()))
    grid_AACE = np.array(list(h2_results_aggregates_grid['Performance By Year']['Annual Average Efficiency [kWh/kg]'].values()))
    ax[2].plot(year_of_operation,offgrid_AACE,color='b',ls='-',lw=2,label='_off-grid')
    ax[2].plot(year_of_operation,grid_AACE,color='r',ls='--',lw=2,label='_grid-connected')
    ax[2].vlines(offgrid_YOR,np.zeros(len(offgrid_YOR)),offgrid_AACE[offgrid_YOR],colors='skyblue',ls='-',lw=0.5,label = 'off-grid stack replacement')
    ax[2].vlines(grid_YOR,np.zeros(len(grid_YOR)),grid_AACE[grid_YOR],colors='tomato',ls='--',lw=0.5,label = 'grid-connected stack replacement')
    ax[2].set_ylabel('Annual Average Efficiency\n[kWh/kg]')
    ax[2].set_xlabel('Year of Operation')
    ax[2].set_xlim((0,plant_life))
    ax[2].set_ylim((0.95*np.min([offgrid_AACE,grid_AACE]),1.05*np.max([offgrid_AACE,grid_AACE])))
    fig.tight_layout()
    []
    


"""
Author: Cory Frontin
Date: 23 Jan 2023
Institution: National Renewable Energy Lab
Description: This file computes von Mises quantities for hemicylindrical tanks,
        replacing Tankinator.xlsx
Sources:
    - Tankinator.xlsx
"""

import numpy as np

def S1(p, Re, R0): # von Mises hoop stress
    return p*(Re**2 + R0**2)/(Re**2 - R0**2)
def S2(p, Re, R0): # von Mises axial stress
    return p*R0**2/(Re**2 - R0**2)
def S3(p, Re, R0): # von Mises radial stress
    return -p

def getPeakStresses(p, Re, R0,
                    proof_factor= 3./2.,
                    burst_factor= 2.25):
    aVM= np.sqrt(2)/2
    bVM= (S2(p, Re, R0) - S1(p, Re, R0))**2
    cVM= (S3(p, Re, R0) - S1(p, Re, R0))**2
    dVM= (S3(p, Re, R0) - S2(p, Re, R0))**2
    eVM= np.sqrt(bVM + cVM + dVM)
    Sproof= proof_factor*aVM*eVM
    Sburst= burst_factor*aVM*eVM
    return (Sproof, Sburst)

def wallThicknessAdjustmentFactor(p, Re, R0, Syield, Sultimate,
                                  proof_factor= 3./2.,
                                  burst_factor= 2.25):
    """
    get factor by which to increase thickness when von Mises stresses exceed
    material yield safety margins
    """
    Sproof, Sburst= getPeakStresses(p, Re, R0,
                                    proof_factor, burst_factor)
    WTAF_proof= Sproof/Syield
    WTAF_burst= Sburst/Sultimate
    WTAF= max(WTAF_proof, WTAF_burst)
    return WTAF

def iterate_thickness(p, R0, thickness_in,
                      Syield, Sultimate,
                      proof_factor= 3./2.,
                      burst_factor= 2.25):
    """
    apply the wall thickness adjustment factor, return it w/ new thickness
    """

    Router= R0 + thickness_in
    WTAF= wallThicknessAdjustmentFactor(p, Router, R0,
                                        Syield, Sultimate,
                                        proof_factor, burst_factor)
    
    return max(1.0, WTAF), max(1.0, WTAF)*thickness_in

def cycle(p, R0, thickness_init,
          Syield, Sultimate,
          proof_factor= 3./2.,
          burst_factor= 2.25,
          max_iter= 10,
          WTAF_tol= 1e-6):
    """
    cycle to find a thickness that satisfies the von Mises criteria
    """

    # compute initial thickness, WTAF
    thickness= thickness_init
    WTAF= wallThicknessAdjustmentFactor(p, R0 + thickness, R0,
                                        Syield, Sultimate,
                                        proof_factor, burst_factor)

    # iterate while WTAF is greater than zero
    n_iter= 0
    while (WTAF - 1.0 > WTAF_tol) and (n_iter < max_iter):

        n_iter += 1 # this cycle iteration number

        # get the next thickness
        WTAF, thickness= iterate_thickness(p, R0, thickness,
                                           Syield, Sultimate,
                                           proof_factor, burst_factor)
    
    return (thickness, WTAF, n_iter)





# from greenheart.hydrogen.electrolysis.PEM_H2_LT_electrolyzer import PEM_electrolyzer_LT
from greenheart.hydrogen.electrolysis.PEM_electrolyzer_IVcurve import PEM_electrolyzer_LT
from numpy.lib.function_base import average
import examples.H2_Analysis.H2AModel as H2AModel
import numpy as np
import pandas as pd


def run_h2_PEM(electrical_generation_timeseries, electrolyzer_size,
                kw_continuous,forced_electrolyzer_cost_kw,lcoe,
                adjusted_installed_cost,useful_life,net_capital_costs,
                voltage_type="constant", stack_input_voltage_DC=250, min_V_cell=1.62,
                p_s_h2_bar=31, stack_input_current_lower_bound=500, cell_active_area=1250, 
                N_cells=130, total_system_electrical_usage=55.5):

    in_dict = dict()
    out_dict = dict()
    in_dict['P_input_external_kW'] = electrical_generation_timeseries
    in_dict['electrolyzer_system_size_MW'] = electrolyzer_size
    el = PEM_electrolyzer_LT(in_dict, out_dict)

    # el.power_supply_rating_MW = electrolyzer_size
    # el.power_supply_rating_MW = power_supply_rating_MW
   #  print("electrolyzer size: ", electrolyzer_size)
   #  el.electrolyzer_system_size_MW = electrolyzer_size
   #  el.input_dict['voltage_type'] = voltage_type
   #  el.stack_input_voltage_DC = stack_input_voltage_DC
   # el.stack_input_voltage_DC = 
    # Assumptions:
   #  el.min_V_cell = min_V_cell  # Only used in variable voltage scenario
   #  el.p_s_h2_bar = p_s_h2_bar   # H2 outlet pressure
   #  el.stack_input_current_lower_bound = stack_input_current_lower_bound
   #  el.cell_active_area = cell_active_area
   #  el.N_cells = N_cells
   #  print("running production rate")
   #  el.h2_production_rate()

    el.h2_production_rate()
    el.water_supply()

    avg_generation = np.mean(electrical_generation_timeseries)  # Avg Generation
    # print("avg_generation: ", avg_generation)
    cap_factor = avg_generation / kw_continuous

    hydrogen_hourly_production = out_dict['h2_produced_kg_hr_system']
    water_hourly_usage = out_dict['water_used_kg_hr']
    water_annual_usage = out_dict['water_used_kg_annual']
    electrolyzer_total_efficiency = out_dict['total_efficiency']
   #  print('water annual: ', water_annual_usage)
    # print("cap_factor: ", cap_factor)

    # Get Daily Hydrogen Production - Add Every 24 hours
    i = 0
    daily_H2_production = []
    while i <= 8760:
        x = sum(hydrogen_hourly_production[i:i + 24])
        daily_H2_production.append(x)
        i = i + 24

    avg_daily_H2_production = np.mean(daily_H2_production)  # kgH2/day
    hydrogen_annual_output = sum(hydrogen_hourly_production)  # kgH2/year
    # elec_remainder_after_h2 = combined_hybrid_curtailment_hopp

    H2A_Results = H2AModel.H2AModel(cap_factor, avg_daily_H2_production, hydrogen_annual_output, force_system_size=True,
                                   forced_system_size=electrolyzer_size, force_electrolyzer_cost=True,
                                    forced_electrolyzer_cost_kw=forced_electrolyzer_cost_kw, useful_life = useful_life)


    feedstock_cost_h2_levelized_hopp = lcoe * total_system_electrical_usage / 100  # $/kg
    # Hybrid Plant - levelized H2 Cost - HOPP
    feedstock_cost_h2_via_net_cap_cost_lifetime_h2_hopp = adjusted_installed_cost / \
                                                          (hydrogen_annual_output * useful_life)  # $/kgH2

    # Total Hydrogen Cost ($/kgH2)
    h2a_costs = H2A_Results['Total Hydrogen Cost ($/kgH2)']
    total_unit_cost_of_hydrogen = h2a_costs + feedstock_cost_h2_levelized_hopp
    feedstock_cost_h2_via_net_cap_cost_lifetime_h2_reopt = net_capital_costs / (
                                (kw_continuous / total_system_electrical_usage) * (8760 * useful_life))

    H2_Results = {'hydrogen_annual_output':
                        hydrogen_annual_output,
                    'feedstock_cost_h2_levelized_hopp':
                       feedstock_cost_h2_levelized_hopp,
                    'feedstock_cost_h2_via_net_cap_cost_lifetime_h2_hopp':
                       feedstock_cost_h2_via_net_cap_cost_lifetime_h2_hopp,
                    'feedstock_cost_h2_via_net_cap_cost_lifetime_h2_reopt':
                       feedstock_cost_h2_via_net_cap_cost_lifetime_h2_reopt,
                    'total_unit_cost_of_hydrogen':
                       total_unit_cost_of_hydrogen,
                    'cap_factor':
                       cap_factor,
                    'hydrogen_hourly_production':
                        hydrogen_hourly_production,
                     'water_hourly_usage':
                     water_hourly_usage,
                     'water_annual_usage':
                     water_annual_usage,
                     'electrolyzer_total_efficiency':
                     electrolyzer_total_efficiency
                   }

    return H2_Results, H2A_Results








from greenheart.simulation.technologies.hydrogen.h2_transport.h2_compression import Compressor
from pytest import approx, raises

# test that we get the results we got when the code was received
class TestH2Compressor():
    p_inlet = 20 # bar
    p_outlet = 68 # bar
    flow_rate_kg_d = 9311
    n_compressors = 2
    
    def test_capex(self):
        comp = Compressor(self.p_outlet, self.flow_rate_kg_d, p_inlet=self.p_inlet, n_compressors=self.n_compressors)
        comp.compressor_power()
        total_capex, total_OM = comp.compressor_costs()
        assert total_capex == 680590.3412708649

    def test_opex(self):
        comp = Compressor(self.p_outlet, self.flow_rate_kg_d, p_inlet=self.p_inlet, n_compressors=self.n_compressors)
        comp.compressor_power()
        total_capex, total_OM = comp.compressor_costs()
        assert total_OM == 200014.00244504173

    def test_system_power(self):
        comp = Compressor(self.p_outlet, self.flow_rate_kg_d, p_inlet=self.p_inlet, n_compressors=self.n_compressors)
        comp.compressor_power()
        _motor_rating, total_system_power = comp.compressor_system_power()
        assert total_system_power == 246.27314443197918

    def test_system_power_report(self):
        """
        H2A Hydrogen Delivery Infrastructure Analysis Models and Conventional Pathway Options Analysis Results
        DE-FG36-05GO15032
        Interim Report
        Nexant, Inc., Air Liquide, Argonne National Laboratory, Chevron Technology Venture, Gas Technology Institute, National Renewable Energy Laboratory, Pacific Northwest National Laboratory, and TIAX LLC
        May 2008
        Table 2-21
        """
        n_compressors = 1
        flow_rate_kg_per_day = 126593.0 #[kg/day]
        p_outlet = 1227.0*0.0689476 # convert from psi to bar
        p_inlet = 265.0*0.0689476 # convert from psi to bar
        sizing_safety_factor = 1.0 # default is 1.1, for a 10% oversizing of the compressor
        comp = Compressor(p_outlet, flow_rate_kg_per_day, p_inlet=p_inlet, n_compressors=n_compressors, sizing_safety_factor=sizing_safety_factor)
        comp.compressor_power()
        _motor_rating, total_system_power = comp.compressor_system_power()
        assert total_system_power == approx(3627.3907562149357)

    def test_max_flow_rate_per_compressor(self):
        p_inlet = 20 # bar
        p_outlet = 68 # bar
        flow_rate_kg_d = 2*5.41*24*60**2
        n_compressors = 2
        with raises(ValueError, match=r".* 5\.4 .*"):
            comp = Compressor(p_outlet,flow_rate_kg_d, p_inlet=p_inlet, n_compressors=n_compressors)
    


"""
Author: Cory Frontin
Date: 23 Jan 2023
Institution: National Renewable Energy Lab
Description: This file computes pressure vessel thickness, replacing Tankinator.xlsx
Sources:
    - Tankinator.xlsx
"""

import numpy as np
import scipy.optimize as opt

from typing import Union
from enum import Enum
import json
import os

from greenheart.simulation.technologies.hydrogen.h2_storage.pressure_vessel import von_mises

class MetalMaterial(object):
    """
    a class for the material properties for metals used in analysis

    :param metal_type: type of metal to use, must be defined in
            material_properties.json
    :type metal_type: str, must be in list of known materials
    :param approx_method: method to approximate inexact table lookups
    :type approx_method: str, must be in list of known methods
    :raises NotImplementedError: if inputs are not found in known data
    """

    def __init__(self,
                 metal_type : str,
                 approx_method : str= "lookup"):
        # make sure settings are correct
        if approx_method not in ["nearest", "lookup", "interp"]:
            raise NotImplementedError("the requested approximation method " \
                                      + "(%s) is not implemented." % approx_method)
        
        with open(os.path.join(os.path.split(__file__)[0],
                  "material_properties.json"), "r") as mmprop_file:
            mmprop= json.load(mmprop_file)
        if metal_type not in mmprop.keys():
            raise NotImplementedError("the requested metal/material (%s) has not" % metal_type \
                                      + "been implemented in material_properties.json.\n" \
                                      + "available types:", mmprop.keys())
        self.mmprop= mmprop[metal_type] # select the relevant material property set

        # stash validated class data
        self.approx_method= approx_method
        self.metal_type= metal_type

        # density and cost per weight
        self.density= self.mmprop['density_kgccm'] # kg/ccm
        self.cost_rate= self.mmprop['costrate_$kg'] # cost per kg

    # nicely package an approximator function        
    def _get_approx_fun(approx_method):
        interp_fun= None
        if approx_method == "nearest":
            def nearest(xq, x, y):
                x= np.asarray(x)
                y= np.asarray(y)
                idx_y= np.argmin(np.abs(x - xq))
                return y[idx_y]
            return nearest
        elif approx_method == "lookup":
            def lookup(xq, x, y):
                x= np.asarray(x)
                y= np.asarray(y)
                idx_yplus= np.argmin(np.abs(x[x < xq] - xq))
                return y[x < xq][idx_yplus]
            return lookup
        elif approx_method == "interp":
            return lambda xq, x, y: np.interp(xq, x, y)
        else:
            raise LookupError("approx method (%s) not found." % approx_method)

    # stash functions to relate yield and ultimate shear (bar) to temp (degC)
    def yield_shear_fun(self, T):
        return MetalMaterial._get_approx_fun(self.approx_method)(T,
                self.mmprop['tables']['yield']['temp_degC'],
                self.mmprop['tables']['yield']['shear_bar'])
    def ultimate_shear_fun(self, T):
        return MetalMaterial._get_approx_fun(self.approx_method)(T,
                self.mmprop['tables']['ultimate']['temp_degC'],
                self.mmprop['tables']['ultimate']['shear_bar'])

class Tank(object):
    """
    a generalized class to size a pressurized gas tank
    assumed to be cylindrical with hemispherical ends

    :param tank_type: type of tank to be used, which can take values I, III, IV
            referring to all-metal, aluminum-lined carbon fiber, and HDPE-lined
            carbon fiber, respectively, as 
    :type tank_type: int, must be 1, 3, or 4
    :param material: material that the pressure vessel is made of
    :type material: str, must be in valid types
    """
    def __init__(self,
                 tank_type : int,
                 yield_factor : float= 3./2.,
                 ultimate_factor : float= 2.25):

        # unpack the key variables
        if not tank_type in [1, 3, 4]:
            raise NotImplementedError("tank_type %d has not been implemented yet.\n" % tank_type)
        self.tank_type= tank_type

        # if not (tank_type == 1):
        #     raise NotImplementedError("haven't done other classes yet. -CVF")
        # self.liner= None

        # store fixed attributes
        self.yield_factor= yield_factor
        self.ultimate_factor= ultimate_factor

        # to start up: undefined geometry values
        self.length_inner= None # inner total length of tank (m)
        self.radius_inner= None # inner radius of tank (m)

        # operating conditions
        self.operating_temperature= None
        self.operating_pressure= None

        self.check_tol= 1e-10 # for validations

    # return functions for symmetry
    def get_length_inner(self):
        return self.length_inner
    def get_radius_inner(self):
        return self.radius_inner
    def get_volume_inner(self):
        """ computes the inner volume """
        return Tank.compute_hemicylinder_volume(self.radius_inner, self.length_inner)
    def get_operating_temperature(self):
        return self.operating_temperature
    def get_operating_pressure(self):
        return self.operating_pressure

    # set functions: specify two of (length, radius, volume)
    def set_length_radius(self, length_in, radius_in):
        """
        set the pressure vessel dimensions by length and radius in cm and
        compute the volume in ccm
        """
        self.length_inner= length_in
        self.radius_inner= radius_in
        self.volume_inner= Tank.compute_hemicylinder_volume(radius_in, length_in)
    def set_length_volume(self, length_in, volume_in):
        """
        set pressure vessel dimensions by length in cm and volume in ccm

        sets the length and volume of the pressure volume, backsolves for the
        radius of the pressure volume
        """
        self.length_inner= length_in
        Rguess= length_in/3
        r_opt= opt.fsolve(lambda x: Tank.compute_hemicylinder_volume(x, length_in) - volume_in, Rguess)
        assert np.abs(Tank.compute_hemicylinder_volume(r_opt, length_in) - volume_in)/volume_in <= self.check_tol
        self.radius_inner= float(r_opt)
    def set_radius_volume(self, radius_in, volume_in):
        """
        set pressure vessel dimensions by radius in cm and volume in ccm

        sets the radius and volume of the pressure volume, backsolves for the
        length of the pressure volume
        """
        self.radius_inner= radius_in
        Lguess= 3*radius_in
        L_opt= opt.fsolve(lambda x: Tank.compute_hemicylinder_volume(radius_in, x) - volume_in, Lguess)
        assert np.abs(Tank.compute_hemicylinder_volume(radius_in, L_opt) - volume_in)/volume_in <= self.check_tol
        self.length_inner= float(L_opt)
    def set_operating_temperature(self, temperature_in):
        self.operating_temperature = temperature_in
    def set_operating_pressure(self, pressure_in):
        self.operating_pressure = pressure_in

    # useful static methods
    def compute_hemicylinder_volume(R : float, L : float) -> float:
        assert (L >= 2*R) # cylindrical tank with hemispherical ends
        return np.pi*R**2*(L - 2.*R/3.)
    def compute_hemicylinder_outer_length(L : float, t : float) -> float:
        return L + 2*t
    def compute_hemicylinder_outer_radius(R : float, t : float) -> float:
        return R + t
    def check_thinwall(Rinner : float, t : float, thinwallratio_crit= 10) -> bool:
        return (Rinner/t >= thinwallratio_crit)

class TypeITank(Tank):
    """
    a class I tank: metal shell tank
    """
   
    def __init__(self,
                 material : str,
                 yield_factor : float= 3./2.,
                 ultimate_factor : float= 2.25,
                 shear_approx = "interp"):

        super().__init__(1, yield_factor, ultimate_factor)
        
        self.material= MetalMaterial(material, approx_method= shear_approx)

        # initial geometry values undefined
        self.thickness= None # thickness of tank

    # return functions for symmetry
    def get_thickness(self):
        return self.thickness
    
    # get the outer dimensions
    def get_length_outer(self):
        """ returns the outer length of the pressure vessel in cm """
        if None in [self.length_inner, self.thickness]:
            return None
        return Tank.compute_hemicylinder_outer_length(self.length_inner, self.thickness)
    def get_radius_outer(self):
        """ returns the outer radius of the pressure vessel in cm """
        if None in [self.radius_inner, self.thickness]:
            return None
        return Tank.compute_hemicylinder_outer_radius(self.radius_inner, self.thickness)
    def get_volume_outer(self):
        """
        returns the outer volume of the pressure vessel in ccm
        """
        if None in [self.length_inner, self.radius_inner, self.thickness]:
            return None
        return Tank.compute_hemicylinder_volume(self.get_radius_outer(), self.get_length_outer())

    def get_volume_metal(self):
        """
        returns the (unsealed) displacement volume of the pressure vessel in ccm
        """
        volume_inner= self.get_volume_inner()
        volume_outer= self.get_volume_outer()
        if None in [volume_inner, volume_outer]: return None
        assert volume_outer >= volume_inner
        return (volume_outer - volume_inner)
    def get_mass_metal(self):
        """ returns the mass of the pressure vessel in kg """
        volume_metal= self.get_volume_metal()
        if volume_metal is None: return None
        return self.material.density*volume_metal
    def get_cost_metal(self):
        """
        returns the cost of the metal in the pressure vessel in dollars
        """
        mass_metal= self.get_mass_metal()
        if mass_metal is None: return None
        return self.material.cost_rate*mass_metal
    
    def get_gravimetric_tank_efficiency(self):
        """
        returns the gravimetric tank efficiency:
        $$ \frac{m_{metal}}{V_{inner}} $$
        in L/kg
        """
        mass_metal= self.get_mass_metal()
        volume_inner= self.get_volume_inner()
        return (volume_inner/1e3)/mass_metal

    def get_yield_thickness(self,
                            pressure : float = None,
                            temperature : float = None):
        """
        gets the yield thickness

        returns the yield thickness given by:
        $$
        t_y= \frac{p R_0}{S_y} \times SF_{yield}
        $$
        with yield safety factor $SF_{yield}= 3/2$ by default

        temperature and pressure must be set in the class, or specified in this
        function

        :param pressure: operating pressure, in bar
        :type pressure: float
        :param temperature: operating temperature, in degrees C
        :type temperature: float
        """

        if (temperature is None) and (self.operating_temperature is None):
            raise LookupError("you must specify an operating temperature.")
        elif temperature is None:
            temperature= self.operating_temperature

        if (pressure is None) and (self.operating_pressure is None):
            raise LookupError("you must specify an operating pressure.")
        elif pressure is None:
            pressure= self.operating_pressure

        Sy= self.material.yield_shear_fun(temperature)

        thickness_yield= pressure*self.radius_inner/Sy*self.yield_factor

        return thickness_yield

    def get_ultimate_thickness(self,
                               pressure : float = None,
                               temperature : float = None):
        """
        get the ultimate thickness

        returns the ultimate thicnkess given by:
        $$
        t_u= \frac{p R_0}{S_u} \times SF_{ultimate}
        $$
        with ultimate safety factor $SF_{yield}= 2.25$ by default

        temperature and pressure must be set in the class, or specified in this
        function

        :param pressure: operating pressure, in bar
        :type pressure: float
        :param temperature: operating temperature, in degrees C
        :type temperature: float
        """

        if (temperature is None) and (self.operating_temperature is None):
            raise LookupError("you must specify an operating temperature.")
        elif temperature is None:
            temperature= self.operating_temperature

        if (pressure is None) and (self.operating_pressure is None):
            raise LookupError("you must specify an operating pressure.")
        elif pressure is None:
            pressure= self.operating_pressure

        Su= self.material.ultimate_shear_fun(temperature)

        thickness_ultimate= pressure*self.radius_inner/Su*self.ultimate_factor

        return thickness_ultimate
    
    def get_thickness_thinwall(self,
                               pressure : float = None,
                               temperature : float = None):
        """
        get the thickness based on thinwall assumptions

        maximum between yield and ultimate thickness

        temperature and pressure must be set in the class, or specified in this
        function

        :param pressure: operating pressure, in bar
        :type pressure: float
        :param temperature: operating temperature, in degrees C
        :type temperature: float
        """

        t_y= self.get_yield_thickness(pressure, temperature)
        t_u= self.get_ultimate_thickness(pressure, temperature)

        thickness= max(t_y, t_u)

        return thickness

    def set_thickness_thinwall(self,
                               pressure : float = None,
                               temperature : float = None):
        """
        set the thickness based on thinwall assumptions

        maximum between yield and ultimate thickness

        temperature and pressure must be set in the class, or specified in this
        function

        :param pressure: operating pressure, in bar
        :type pressure: float
        :param temperature: operating temperature, in degrees C
        :type temperature: float
        """

        self.thickness= self.get_thickness_thinwall(pressure, temperature)

    def get_thickness_vonmises(self,
                               pressure : float = None,
                               temperature : float = None,
                               max_cycle_iter : int = 10,
                               adj_fac_tol : float = 1e-6):
        """
        get the thickness based on a von Mises cycle

        temperature and pressure must be set in the class, or specified here

        :param pressure: operating pressure, in bar
        :type pressure: float
        :param temperature: operating temperature, in degrees C
        :type temperature: float
        :param max_cycle_iter: maximum iterations for von Mises cycle
        :type max_cycle_iter: int
        :param adj_fac_tol: tolerance for close enough wall thickness adjustment
                factor
        """

        if (temperature is None) and (self.operating_temperature is None):
            raise LookupError("you must specify an operating temperature.")
        elif temperature is None:
            temperature= self.operating_temperature

        if (pressure is None) and (self.operating_pressure is None):
            raise LookupError("you must specify an operating pressure.")
        elif pressure is None:
            pressure= self.operating_pressure

        # get the limit shears
        Sy= self.material.yield_shear_fun(temperature)
        Su= self.material.ultimate_shear_fun(temperature)

        # start from the thinwall analysis
        thickness_init= self.get_thickness_thinwall(pressure, temperature)

        # check to see if von Mises analysis is even needed
        if (Tank.check_thinwall(self.radius_inner, thickness_init)) \
                and (von_mises.wallThicknessAdjustmentFactor(pressure,
                                                             self.radius_inner + thickness_init,
                                                             self.radius_inner,
                                                             Sy, Su) == 1.0):
            thickness_cycle= thickness_init # trivially satisfied
            iter_cycle= -1
            print("trivially satisfied")
        else:
            print("running von mises cycle")
            print(pressure, self.radius_inner, thickness_init, Sy, Su)
            (thickness_cycle, WTAF_cycle, iter_cycle)= \
                    von_mises.cycle(pressure, self.radius_inner, thickness_init,
                                    Sy, Su,
                                    max_iter= max_cycle_iter,
                                    WTAF_tol= adj_fac_tol)
        return thickness_cycle, iter_cycle
    def set_thickness_vonmises(self,
                               pressure : float = None,
                               temperature : float = None,
                               max_cycle_iter : int = 10,
                               adj_fac_tol : float = 1e-6):
        """
        set the thickness based on a von Mises cycle

        temperature and pressure must be set in the class, or specified here

        :param pressure: operating pressure, in bar
        :type pressure: float
        :param temperature: operating temperature, in degrees C
        :type temperature: float
        :param max_cycle_iter: maximum iterations for von Mises cycle
        :type max_cycle_iter: int
        :param adj_fac_tol: tolerance for close enough wall thickness adjustment
                factor
        """

        thickness, iter= self.get_thickness_vonmises(pressure, temperature,
                                                     max_cycle_iter, adj_fac_tol)
        self.thickness= thickness

    

class LinedTank(Tank):
    """
    a lined tank for Type III or Type III: aluminum-lined carbon fiber-jacketed tank
    """

    def __init__(self,
                 tanktype : int,
                 load_bearing_liner,
                 liner_design_load_factor= 0.21,
                 liner_thickness_min = 0.3,
                 yield_factor : float= 3./2.,
                 ultimate_factor : float= 2.25):
        super().__init__(tanktype, yield_factor, ultimate_factor)

        if tanktype not in [3, 4]:
            raise NotImplementedError("unknown tank type.")

        self.load_bearing_liner= load_bearing_liner
        self.liner_design_load_factor= liner_design_load_factor

        with open(os.path.join(os.path.split(__file__)[0],
                  "material_properties.json"), "r") as mmprop_file:
            mmprop= json.load(mmprop_file)

        # liner characteristics
        if tanktype == 3:
            self.shear_ultimate_liner= mmprop["liner_aluminum"]["shear_ultimate_bar"]
            self.density_liner= mmprop["liner_aluminum"]["density_kgccm"]
            self.costrate_liner= mmprop["liner_aluminum"]["costrate_$kg"]
        else:
            self.shear_ultimate_liner= None
            self.density_liner= mmprop["HDPE"]["density_kgccm"]
            self.costrate_liner= mmprop["HDPE"]["costrate_$kg"]
        self.thickness_liner_min= liner_thickness_min
        
        # jacket characteristics
        self.shear_ultimate_jacket= mmprop["carbon_fiber"]["shear_ultimate_bar"]
        self.density_jacket= mmprop["carbon_fiber"]["density_kgccm"]
        self.costrate_jacket= mmprop["carbon_fiber"]["costrate_$kg"]
        self.fiber_translation_efficiency_jacket= mmprop["carbon_fiber"]["fiber_translation_efficiency"]
        self.fiber_layer_thickness= mmprop["carbon_fiber"]["layer_thickness_cm"]
        self.fiber_layer_min= mmprop["carbon_fiber"]["min_layer"]

        # thicknesses (to be computed)
        self.thickness_liner= None
        self.thickness_jacket= None
        self.thickness_ideal_jacket= None
        self.Nlayer_jacket= None

    def get_thicknesses_thinwall(self, pressure : float = None):
        """ ??? """

        if (pressure is None) and (self.operating_pressure is None):
            raise LookupError("you must specify an operating pressure.")
        elif pressure is None:
            pressure= self.operating_pressure

        # compute the liner thickness
        pressure_burst_target= self.ultimate_factor*pressure
        if self.tank_type != 4:
            pressure_liner_target= pressure_burst_target*self.liner_design_load_factor
            thickness_burst= pressure_liner_target*self.radius_inner/self.shear_ultimate_liner
            thickness_liner= max(thickness_burst, self.thickness_liner_min)
        else:
            thickness_liner= self.thickness_liner_min

        # compute ideal jacket thickness
        radius_liner= Tank.compute_hemicylinder_outer_radius(self.radius_inner, thickness_liner)
        thickness_jacket_ideal= pressure_burst_target*radius_liner \
                /self.shear_ultimate_jacket/self.fiber_translation_efficiency_jacket
        if self.load_bearing_liner:
            # subdivide pressure if liner is load-bearing
            pressure_liner= thickness_liner*self.shear_ultimate_liner/self.radius_inner
            pressure_jacket= pressure_burst_target - pressure_liner
            assert pressure_jacket >= 0
            thickness_jacket_ideal= pressure_jacket*radius_liner/self.shear_ultimate_jacket \
                    /self.fiber_translation_efficiency_jacket

        # compute number of layers, real thickness
        Nlayer_jacket= max(self.fiber_layer_min,
                           int(np.ceil(thickness_jacket_ideal/self.fiber_layer_thickness)))
        thickness_jacket_real= Nlayer_jacket*self.fiber_layer_thickness

        return thickness_liner, thickness_jacket_ideal, Nlayer_jacket, thickness_jacket_real

    def set_thicknesses_thinwall(self, pressure : float = None):
        """ ??? """

        if (pressure is None) and (self.operating_pressure is None):
            raise LookupError("you must specify an operating pressure.")
        elif pressure is None:
            pressure= self.operating_pressure

        # pass the returns of the previous function
        self.thickness_liner, self.thickness_ideal_jacket, \
                self.Nlayer_jacket, self.thickness_jacket= \
                self.get_thicknesses_thinwall(pressure)

    def get_safetyfactor_real_jacket(self):
        """
        figure out the integer layer safety factor
        """

        if None in [self.thickness_jacket, self.thickness_ideal_jacket]: return None

        sf_real= self.ultimate_factor*self.thickness_jacket/self.thickness_jacket_ideal
        
        return sf_real

    # get the liner dimensions
    def get_length_liner(self):
        """ returns the outer length of the pressure vessel in cm """
        if None in [self.length_inner, self.thickness_liner]:
            return None
        return Tank.compute_hemicylinder_outer_length(self.length_inner, self.thickness_liner)
    def get_radius_liner(self):
        """ returns the outer radius of the pressure vessel in cm """
        if None in [self.radius_inner, self.thickness_liner]:
            return None
        return Tank.compute_hemicylinder_outer_radius(self.radius_inner, self.thickness_liner)
    def get_volume_outer_liner(self):
        """
        returns the outer volume of the pressure vessel in ccm
        """
        if None in [self.length_inner, self.radius_inner, self.thickness_liner]:
            return None
        return Tank.compute_hemicylinder_volume(self.get_radius_liner(), self.get_length_liner())
    def get_volume_liner(self):
        """
        returns the (unsealed) displacement volume of the liner in ccm
        """
        volume_inner= self.get_volume_inner()
        volume_outer_liner= self.get_volume_outer_liner()
        if None in [volume_inner, volume_outer_liner]: return None
        assert volume_outer_liner >= volume_inner
        return (volume_outer_liner - volume_inner)
    def get_mass_liner(self):
        """
        returns the mass of the liner in kg
        """
        volume_liner= self.get_volume_liner()
        if volume_liner is None: return None
        return self.density_liner*volume_liner
    def get_cost_liner(self):
        """
        returns the cost of the liner material in $
        """
        mass_liner= self.get_mass_liner()
        if mass_liner is None: return None
        return self.costrate_liner*mass_liner

    # get the outer dimensions
    def get_length_outer(self):
        """ returns the outer length of the pressure vessel in cm """
        if None in [self.length_inner, self.thickness_liner, self.thickness_jacket]:
            return None
        return Tank.compute_hemicylinder_outer_length(self.length_inner,
                self.thickness_liner + self.thickness_jacket)
    def get_radius_outer(self):
        """ returns the outer radius of the pressure vessel in cm """
        if None in [self.radius_inner, self.thickness_liner, self.thickness_jacket]:
            return None
        return Tank.compute_hemicylinder_outer_radius(self.radius_inner,
                self.thickness_liner + self.thickness_jacket)
    def get_volume_outer(self):
        """
        returns the outer volume of the pressure vessel in ccm
        """
        if None in [self.length_inner, self.radius_inner, self.thickness_liner, self.thickness_jacket]:
            return None
        return Tank.compute_hemicylinder_volume(self.get_radius_outer(), self.get_length_outer())
    def get_volume_jacket(self):
        """
        returns the (unsealed) displacement volume of the carbon fiber jacket in ccm
        """
        volume_outer= self.get_volume_outer()
        volume_outer_liner= self.get_volume_outer_liner()
        if None in [volume_outer, volume_outer_liner]: return None
        assert volume_outer >= volume_outer_liner
        return (volume_outer - volume_outer_liner)
    def get_mass_jacket(self):
        """
        returns the mass of the carbon fiber jacket in kg
        """
        volume_jacket= self.get_volume_jacket()
        if volume_jacket is None: return None
        return self.density_jacket*volume_jacket
    def get_cost_jacket(self):
        """
        returns the cost of the jacket material in $
        """
        mass_jacket= self.get_mass_jacket()
        if mass_jacket is None: return None
        return self.costrate_jacket*mass_jacket
    
    def get_mass_tank(self):
        """
        returns the mass of the empty tank in kg
        """
        mass_liner= self.get_mass_liner()
        mass_jacket= self.get_mass_jacket()
        if None in [mass_liner, mass_jacket]: return None
        return mass_liner + mass_jacket
    def get_cost_tank(self):
        """
        returns the material cost of the tank in $
        """
        cost_liner= self.get_cost_liner()
        cost_jacket= self.get_cost_jacket()
        if None in [cost_liner, cost_jacket]: return None
        return cost_liner + cost_jacket
    def get_gravimetric_tank_efficiency(self):
        """
        returns the gravimetric tank efficiency:
        $$ \frac{m_{tank}}{V_{inner}} $$
        in L/kg
        """
        mass_tank= self.get_mass_tank()
        volume_inner= self.get_volume_inner()
        return (volume_inner/1e3)/mass_tank

class TypeIIITank(LinedTank):
    def __init__(self,
                 conservative= False,
                 liner_design_load_factor= 0.21,
                 liner_thickness_min= 0.3,
                 yield_factor: float = 3 / 2,
                 ultimate_factor: float = 2.25):
                load_bearing_liner= not conservative # use load bearing liner iff not conservative
                super().__init__(3, load_bearing_liner, liner_design_load_factor,
                         liner_thickness_min, yield_factor, ultimate_factor)

class TypeIVTank(LinedTank):
    def __init__(self,
                 liner_design_load_factor= 0.21,
                 liner_thickness= 0.4,
                 yield_factor: float = 3/2,
                 ultimate_factor: float = 2.25):
        super().__init__(4, False, liner_design_load_factor,
                         liner_thickness, yield_factor, ultimate_factor)




from pytest import approx

from greenheart.simulation.technologies.hydrogen.desal.desal_model_eco import RO_desal_eco

# Test values are based on hand calculations

class TestRO_desal():
    rel_tol = 1E-2

    freshwater_needed = 10000   #[kg/hr]

    saltwater = RO_desal_eco(freshwater_needed, "Seawater")
    brackish = RO_desal_eco(freshwater_needed, "Brackish")

    def test_capacity_m3_per_hr(self):
        assert self.saltwater[0] == approx(10.03, rel=1E-5)
        assert self.brackish[0] == approx(10.03, rel=1E-5)

    def test_feedwater(self):
        assert self.saltwater[1] == approx(20.06, rel=1E-5)
        assert self.brackish[1] == approx(13.37, rel=1E-3)

    def test_power(self):
        assert self.saltwater[2] == approx(40.12, rel=1E-5)
        assert self.brackish[2] == approx(15.04, rel=1E-3)

    def test_capex(self):
        assert self.saltwater[3] == approx(91372, rel=1E-2)
        assert self.brackish[3] == approx(91372, rel=1E-2)
    
    def test_opex(self):
        assert self.saltwater[4] == approx(13447, rel=1E-2)
        assert self.brackish[4] == approx(13447, rel=1E-2)

    def test_RO_Desal_Seawater(self):
        '''Test Seawater RO Model'''
        outputs=RO_desal_eco(freshwater_kg_per_hr=997,salinity='Seawater')
        RO_desal_mass = outputs[5]
        RO_desal_footprint = outputs[6]
        assert approx(RO_desal_mass) == 346.7
        assert approx(RO_desal_footprint) == .467

    def test_RO_Desal_distributed(self):
        '''Test Seawater RO Model'''
        n_systems = 2
        total_freshwater_kg_per_hr_required = 997
        per_system_freshwater_kg_per_hr_required = total_freshwater_kg_per_hr_required/n_systems

        total_outputs=RO_desal_eco(freshwater_kg_per_hr=total_freshwater_kg_per_hr_required,salinity='Seawater')
        per_system_outputs = RO_desal_eco(per_system_freshwater_kg_per_hr_required, salinity="Seawater")

        for t, s in zip(total_outputs, per_system_outputs):
            assert t == approx(s*n_systems)

    
    def test_RO_Desal_Brackish(self):
        '''Test Brackish Model'''
        outputs=RO_desal_eco(freshwater_kg_per_hr=997,salinity='Brackish')
        RO_desal_mass = outputs[5]
        RO_desal_footprint = outputs[6]
        assert approx(RO_desal_mass) == 346.7
        assert approx(RO_desal_footprint) == .467
        
if __name__ == "__main__":
    test_set = TestRO_desal()


from greenheart.simulation.technologies.hydrogen.electrolysis.PEM_costs_Singlitico_model import PEMCostsSingliticoModel
from greenheart.simulation.technologies.hydrogen.electrolysis.PEM_electrolyzer_IVcurve import PEM_electrolyzer_LT
from greenheart.simulation.technologies.hydrogen.electrolysis.PEM_H2_LT_electrolyzer import PEM_electrolyzer_LT
import greenheart.simulation.technologies.hydrogen.electrolysis.PEM_H2_LT_electrolyzer_Clusters
import greenheart.simulation.technologies.hydrogen.electrolysis.PEM_tools


import numpy as np
import os
from hopp.simulation.technologies import REopt
from hopp.simulation.technologies.pv_source import PVPlant
from hopp.simulation.technologies.wind_source import WindPlant
import PySAM.Singleowner as so
import pandas as pd
import pickle
# import post_and_poll
from dotenv import load_dotenv
import json

#TODO:
# - Change to developer API.
# - Integrate all new features of post_and_poll with REOpt class (from hybrid/reopt_dev)


def run_reopt(site, scenario, load, interconnection_limit_kw, critical_load_factor, useful_life,
              battery_can_grid_charge,
              storage_used, run_reopt_flag):

    # kw_continuous = forced_system_size  # 5 MW continuous load - equivalent to 909kg H2 per hr at 55 kWh/kg electrical intensity

    urdb_label = "5ca4d1175457a39b23b3d45e"  # https://openei.org/apps/IURDB/rate/view/5ca3d45ab718b30e03405898
    pv_config = {'system_capacity_kw': 20000}
    solar_model = PVPlant(site, pv_config)
    # ('num_turbines', 'turbine_rating_kw', 'rotor_diameter', 'hub_height', 'layout_mode', 'layout_params')
    wind_config = {'num_turbines': np.floor(scenario['Wind Size MW'] / scenario['Turbine Rating']),
                       'rotor_dimeter': scenario['Rotor Diameter'], 'hub_height': scenario['Tower Height'],
                   'turbine_rating_kw': scenario['Turbine Rating']}

    wind_model = WindPlant(site, wind_config)
    fin_model = so.default("GenericSystemSingleOwner")
    filepath = os.path.dirname(os.path.abspath(__file__))
    fileout = 'reopt_result_test_intergration.json'
    # site = SiteInfo(sample_site, hub_height=tower_height)
    count = 1
    reopt = REopt(lat=scenario['Lat'],
                  lon=scenario['Long'],
                  load_profile=load,
                  urdb_label=urdb_label,
                  solar_model=solar_model,
                  wind_model=wind_model,
                  fin_model=fin_model,
                  interconnection_limit_kw=interconnection_limit_kw,
                  off_grid=True,
                  fileout=fileout)

    reopt.set_rate_path(os.path.join(filepath, '../data'))

    reopt.post['Scenario']['Site']['Wind']['installed_cost_us_dollars_per_kw'] = scenario['Wind Cost KW']  # ATB
    reopt.post['Scenario']['Site']['PV']['installed_cost_us_dollars_per_kw'] = scenario['Solar Cost KW']
    reopt.post['Scenario']['Site']['Storage'] = {'min_kw': 0.0, 'max_kw': 0.99e9, 'min_kwh': 0.0,
                                                 'max_kwh': 0.99e9,
                                                 'internal_efficiency_pct': 0.975, 'inverter_efficiency_pct': 0.96,
                                                 'rectifier_efficiency_pct': 0.96, 'soc_min_pct': 0.2,
                                                 'soc_init_pct': 0.5,
                                                 'canGridCharge': battery_can_grid_charge,
                                                 'installed_cost_us_dollars_per_kw': scenario['Storage Cost KW'],
                                                 'installed_cost_us_dollars_per_kwh': scenario['Storage Cost KWh'],
                                                 'replace_cost_us_dollars_per_kw': scenario['Storage Cost KW'],
                                                 'replace_cost_us_dollars_per_kwh': scenario['Storage Cost KWh'],
                                                 'inverter_replacement_year': 10,
                                                 'battery_replacement_year': 10, 'macrs_option_years': 7,
                                                 'macrs_bonus_pct': 1.0, 'macrs_itc_reduction': 0.5,
                                                 'total_itc_pct': 0.0,
                                                 'total_rebate_us_dollars_per_kw': 0,
                                                 'total_rebate_us_dollars_per_kwh': 0}

    reopt.post['Scenario']['Site']['Financial']['analysis_years'] = useful_life
    if not storage_used:
        reopt.post['Scenario']['Site']['Storage']['max_kw'] = 0
    if scenario['PTC Available']:
        reopt.post['Scenario']['Site']['Wind']['pbi_us_dollars_per_kwh'] = 0.022
    else:
        reopt.post['Scenario']['Site']['Wind']['pbi_us_dollars_per_kwh'] = 0.0
    if scenario['ITC Available']:
        reopt.post['Scenario']['Site']['PV']['federal_itc_pct'] = 0.26
    else:
        reopt.post['Scenario']['Site']['PV']['federal_itc_pct'] = 0.0

    # reopt.post['Scenario']['Site']['LoadProfile']['doe_reference_name'] = "FlatLoad"
    # reopt.post['Scenario']['Site']['LoadProfile']['annual_kwh'] = load #8760 * kw_continuous
    reopt.post['Scenario']['Site']['LoadProfile']['loads_kw'] = load
    reopt.post['Scenario']['Site']['LoadProfile']['critical_load_pct'] = critical_load_factor

    off_grid = False
    reopt.post['Scenario']['optimality_tolerance_techs'] = 0.05

    if off_grid == True:
        # reopt.post['Scenario']['Site'].pop('Wind')
        # reopt.post['Scenario']['Site']['Wind']['min_kw'] = 10000
        dictforstuff = {"off_grid_flag": True}
        reopt.post['Scenario'].update(dictforstuff)
        reopt.post['Scenario']['optimality_tolerance_techs'] = 0.05
        reopt.post['Scenario']["timeout_seconds"] = 3600
        # reopt.post['Scenario']['Site']['LoadProfile'].pop('annual kwh')
        reopt.post['Scenario']['Site'].pop('ElectricTariff')
        reopt.post['Scenario']['Site']['LoadProfile']['critical_load_pct'] = 1.0
        f = open('massproducer_offgrid (1).json')
        data_for_post = json.load(f)
        reopt.post['Scenario']['Site']['Financial'] = data_for_post['Scenario']['Site']['Financial']
    else:
        reopt.post['Scenario']['Site']['ElectricTariff']['wholesale_rate_us_dollars_per_kwh'] = 0.01
        reopt.post['Scenario']['Site']['ElectricTariff']['wholesale_rate_above_site_load_us_dollars_per_kwh'] = 0.01
        reopt.post['Scenario']['Site']['LoadProfile']['outage_start_hour'] = 10
        reopt.post['Scenario']['Site']['LoadProfile']['outage_end_hour'] = 11

    from pathlib import Path
    post_path = 'results/reopt_precomputes/reopt_post'
    post_path_abs = Path(__file__).parent / post_path
    if not os.path.exists(post_path_abs.parent):
        os.mkdir(post_path_abs.parent)
    with open(post_path_abs, 'w') as outfile:
        json.dump(reopt.post, outfile)
    # mass_producer_dict = {
    #     "mass_units": "kg",
    #     "time_units": "hr",
    #     "min_mass_per_time": 10.0,
    #     "max_mass_per_time": 10.0,
    #     "electric_consumed_to_mass_produced_ratio_kwh_per_mass": 71.7,
    #     "thermal_consumed_to_mass_produced_ratio_kwh_per_mass": 0.0,
    #     "feedstock_consumed_to_mass_produced_ratio": 0.0,
    #     "installed_cost_us_dollars_per_mass_per_time": 10.0,
    #     "om_cost_us_dollars_per_mass_per_time": 1.5,
    #     "om_cost_us_dollars_per_mass": 0.0,
    #     "mass_value_us_dollars_per_mass": 5.0,
    #     "feedstock_cost_us_dollars_per_mass": 0.0,
    #     "macrs_option_years": 0,
    #     "macrs_bonus_pct": 0
    # }
    # reopt.post['Scenario']['Site']['MassProducer'] = mass_producer_dict

    if run_reopt_flag:
        #NEW METHOD
        load_dotenv()
        result = reopt.get_reopt_results()

        #BASIC INITIAL TEST FOR NEW METHOD
        # result = post_and_poll.get_api_results(data_for_post, NREL_API_KEY, 'https://offgrid-electrolyzer-reopt-dev-api.its.nrel.gov/v1',
        #                       'reopt_result_test_intergration.json')

        # f = open('massproducer_offgrid (1).json')
        # data_for_post = json.load(f)

        #OLD METHOD
        # result = reopt.get_reopt_results(force_download=True)


        pickle.dump(result, open("results/reopt_precomputes/results_{}_{}_{}.p".format(
            scenario['Site Name'], scenario['Scenario Name'], critical_load_factor), "wb"))

    else:
        print("Not running reopt. Loading Dummy data")
        precompute_path = 'results/reopt_precomputes/'
        precompute_path_abs = Path(__file__).parent / precompute_path
        result = pickle.load(
            open(os.path.join(precompute_path_abs, "results_ATB_moderate_2020_IOWA_0.9.p"), "rb"))

    if result['outputs']['Scenario']['Site']['PV']['size_kw']:
        solar_size_mw = result['outputs']['Scenario']['Site']['PV']['size_kw'] / 1000

    if result['outputs']['Scenario']['Site']['Wind']['size_kw']:
        wind_size_mw = result['outputs']['Scenario']['Site']['Wind']['size_kw'] / 1000

    if result['outputs']['Scenario']['Site']['Storage']['size_kw']:
        storage_size_mw = result['outputs']['Scenario']['Site']['Storage']['size_kw'] / 1000
        storage_size_mwh = result['outputs']['Scenario']['Site']['Storage']['size_kwh'] / 1000
        storage_hours = storage_size_mwh / storage_size_mw

    reopt_site_result = result['outputs']['Scenario']['Site']
    generated_date = pd.date_range(start='1/1/2018 00:00:00', end='12/31/2018 23:00:00', periods=8760)
    if reopt_site_result['Wind']['size_kw'] == 0:

        reopt_site_result['Wind']['year_one_power_production_series_kw'] = np.zeros(8760)
        reopt_site_result['Wind']['year_one_to_grid_series_kw'] = np.zeros(8760)
        reopt_site_result['Wind']['year_one_to_load_series_kw'] = np.zeros(8760)
        reopt_site_result['Wind']['year_one_to_battery_series_kw'] = np.zeros(8760)
        reopt_site_result['Wind']['year_one_curtailed_production_series_kw'] = np.zeros(8760)
        wind_size_mw = 0

    if reopt_site_result['PV']['size_kw'] == 0:
        reopt_site_result['PV']['year_one_power_production_series_kw'] = np.zeros(8760)
        reopt_site_result['PV']['year_one_to_grid_series_kw'] = np.zeros(8760)
        reopt_site_result['PV']['year_one_to_load_series_kw'] = np.zeros(8760)
        reopt_site_result['PV']['year_one_to_battery_series_kw'] = np.zeros(8760)
        reopt_site_result['PV']['year_one_curtailed_production_series_kw'] = np.zeros(8760)
        solar_size_mw = 0

    if reopt_site_result['Storage']['size_kw'] == 0:
        reopt_site_result['Storage']['year_one_soc_series_pct'] = np.zeros(8760)
        reopt_site_result['Storage']['year_one_to_massproducer_series_kw'] = np.zeros(8760)
        storage_size_mw = 0
        storage_size_mwh = 0
        storage_hours = 0

    combined_pv_wind_power_production = [x + y for x, y in
                                         zip(reopt_site_result['PV']['year_one_power_production_series_kw']
                                             , reopt_site_result['Wind']['year_one_power_production_series_kw'])]
    combined_pv_wind_storage_power_production = [x + y for x, y in zip(combined_pv_wind_power_production,
                                                                       reopt_site_result['Storage'][
                                                                           'year_one_to_load_series_kw'])]
    energy_shortfall = [y - x for x, y in zip(combined_pv_wind_storage_power_production, load)]
    energy_shortfall = [x if x > 0 else 0 for x in energy_shortfall]

    combined_pv_wind_curtailment = [x + y for x, y in
                                    zip(reopt_site_result['PV']['year_one_curtailed_production_series_kw']
                                        , reopt_site_result['Wind']['year_one_curtailed_production_series_kw'])]


    reopt_result_dict = {'Date':
                             generated_date,
                         'pv_power_production':
                             reopt_site_result['PV']
                             ['year_one_power_production_series_kw'],
                         'pv_power_to_grid':
                             reopt_site_result['PV']
                             ['year_one_to_grid_series_kw'],
                         'pv_power_to_load':
                             reopt_site_result['PV']['year_one_to_load_series_kw'],
                         'pv_power_to_battery':
                             reopt_site_result['PV']['year_one_to_battery_series_kw'],
                         'pv_power_curtailed':
                             reopt_site_result['PV']['year_one_curtailed_production_series_kw'],
                         'wind_power_production':
                             reopt_site_result['Wind']
                             ['year_one_power_production_series_kw'],
                         'wind_power_to_grid':
                             reopt_site_result['Wind']
                             ['year_one_to_grid_series_kw'],
                         'wind_power_to_load':
                             reopt_site_result['Wind']['year_one_to_load_series_kw'],
                         'wind_power_to_battery':
                             reopt_site_result['Wind']['year_one_to_battery_series_kw'],
                         'wind_power_curtailed':
                             reopt_site_result['Wind']['year_one_curtailed_production_series_kw'],
                         'combined_pv_wind_power_production':
                             combined_pv_wind_power_production,
                         'combined_pv_wind_storage_power_production':
                             combined_pv_wind_storage_power_production,
                         'storage_power_to_load':
                             reopt_site_result['Storage']['year_one_to_load_series_kw'],
                         'storage_power_to_grid':
                             reopt_site_result['Storage']['year_one_to_grid_series_kw'],
                         'battery_soc_pct':
                             reopt_site_result['Storage']['year_one_soc_series_pct'],
                         'energy_shortfall':
                             energy_shortfall,
                         'combined_pv_wind_curtailment':
                             combined_pv_wind_curtailment
                         }

    REoptResultsDF = pd.DataFrame(reopt_result_dict)

    return wind_size_mw, solar_size_mw, storage_size_mw, storage_size_mwh, storage_hours, result, REoptResultsDF


# -*- coding: utf-8 -*-
"""
Created on Fri Dec  2 12:09:20 2022

@author: ereznic2
"""
import pandas as pd

dircambium = 'examples/H2_Analysis/Cambium_data/StdScen21_MidCase95by2035_hourly_' 

# grid_connection_scenario = 'hybrid-grid'
# atb_year = 2020
# site_name = 'TX'
# turbine_model = '6MW'
# electrolysis_scale = 'Centralized'
# policy_option = 'no policy'
# grid_price_scenario = 'retail-flat'
# electrolyzer_energy_kWh_per_kg = 55

def hydrogen_LCA_singlescenario(grid_connection_scenario,atb_year,site_name,turbine_model,electrolysis_scale,policy_option,grid_price_scenario,electrolyzer_energy_kWh_per_kg,hydrogen_hourly_results_RODeO):

    #==============================================================================
    # DATA
    #==============================================================================        
    # Conversions
    g_to_kg_conv  = 0.001  # Conversion from grams to kilograms
    kg_to_MT_conv = 0.001 # Converion from kg to metric tonnes
    MT_to_kg_conv = 1000 # Conversion from metric tonne to kilogram
    kWh_to_MWh_conv = 0.001 # Conversion from kWh to MWh
    
    #------------------------------------------------------------------------------
    # Renewable infrastructure embedded emission intensities
    #------------------------------------------------------------------------------
    system_life        = 30
    ely_stack_capex_EI = 0.019 # PEM electrolyzer CAPEX emissions (kg CO2e/kg H2)
    wind_capex_EI      = 10    # Electricity generation from wind, nominal value taken (g CO2e/kWh)
    solar_pv_capex_EI  = 37    # Electricity generation from solar pv, nominal value taken (g CO2e/kWh)

    
    #------------------------------------------------------------------------------
    # Hydrogen production via water electrolysis
    #------------------------------------------------------------------------------
    
    grid_trans_losses   = 0.05 # Grid losses of 5% are assumed (-)
    fuel_to_grid_curr   = 48   # Fuel mix emission intensity for current power grid (g CO2e/kWh)
    fuel_to_grid_futu   = 14   # Fuel mix emission intensity for future power grid (g CO2e/kWh)
    
    if atb_year == 2020:
        cambium_year = 2025
    elif atb_year == 2025:
        cambium_year = 2030
    elif atb_year == 2030:
        cambium_year =2035
    elif atb_year == 2035:
        cambium_year = 2040
        
    # Read in Cambium data  
    cambiumdata_filepath = dircambium + site_name + '_'+str(cambium_year) + '.csv'
    cambium_data = pd.read_csv(cambiumdata_filepath,index_col = None,header = 4,usecols = ['lrmer_co2_c','lrmer_ch4_c','lrmer_n2o_c','lrmer_co2_p','lrmer_ch4_p','lrmer_n2o_p','lrmer_co2e_c','lrmer_co2e_p','lrmer_co2e'])
    
    cambium_data = cambium_data.reset_index().rename(columns = {'index':'Interval','lrmer_co2_c':'LRMER CO2 combustion (kg-CO2/MWh)','lrmer_ch4_c':'LRMER CH4 combustion (g-CH4/MWh)','lrmer_n2o_c':'LRMER N2O combustion (g-N2O/MWh)',\
                                                  'lrmer_co2_p':'LRMER CO2 production (kg-CO2/MWh)','lrmer_ch4_p':'LRMER CH4 production (g-CH4/MWh)','lrmer_n2o_p':'LRMER N2O production (g-N2O/MWh)','lrmer_co2e_c':'LRMER CO2 equiv. combustion (kg-CO2e/MWh)',\
                                                  'lrmer_co2e_p':'LRMER CO2 equiv. production (kg-CO2e/MWh)','lrmer_co2e':'LRMER CO2 equiv. total (kg-CO2e/MWh)'})
    
    cambium_data['Interval']=cambium_data['Interval']+1
    cambium_data = cambium_data.set_index('Interval')   
    
    # Read in rodeo data
    rodeo_data = hydrogen_hourly_results_RODeO[['Interval','Input Power (MW)','Non-Ren Import (MW)','Renewable Input (MW)','Curtailment (MW)','Product Sold (units of product)']]
    rodeo_data = rodeo_data.rename(columns = {'Input Power (MW)':'Electrolyzer Power (MW)','Non-Ren Import (MW)':'Grid Import (MW)','Renewable Input (MW)':'Renewable Input (MW)', 'Curtailment (MW)':'Curtailment (MW)','Product Sold (units of product)':'Hydrogen production (kg-H2)'})
    # Combine RODeO and Cambium data into one dataframe
    combined_data = rodeo_data.merge(cambium_data, on = 'Interval',how = 'outer')   
    
    # Calculate hourly grid emissions factors of interest. If we want to use different GWPs, we can do that here. The Grid Import is an hourly data i.e., in MWh
    combined_data['Total grid emissions (kg-CO2e)'] = combined_data['Grid Import (MW)']*combined_data['LRMER CO2 equiv. total (kg-CO2e/MWh)']
    combined_data['Scope 2 (combustion) grid emissions (kg-CO2e)'] = combined_data['Grid Import (MW)']*combined_data['LRMER CO2 equiv. combustion (kg-CO2e/MWh)']
    combined_data['Scope 3 (production) grid emissions (kg-CO2e)'] = combined_data['Grid Import (MW)']*combined_data['LRMER CO2 equiv. production (kg-CO2e/MWh)']
    
    # Sum total emissions
    scope2_grid_emissions_sum = combined_data['Scope 2 (combustion) grid emissions (kg-CO2e)'].sum()*system_life*kg_to_MT_conv
    scope3_grid_emissions_sum = combined_data['Scope 3 (production) grid emissions (kg-CO2e)'].sum()*system_life*kg_to_MT_conv
    h2prod_sum = combined_data['Hydrogen production (kg-H2)'].sum()*system_life*kg_to_MT_conv
    h2prod_grid_frac = combined_data['Grid Import (MW)'].sum() / combined_data['Electrolyzer Power (MW)'].sum()
           
    if grid_connection_scenario == 'hybrid-grid' :
        # Calculate grid-connected electrolysis emissions/ future cases should reflect targeted electrolyzer electricity usage
        electrolysis_Scope3_EI =  h2prod_grid_frac*scope3_grid_emissions_sum/h2prod_sum + wind_capex_EI * electrolyzer_energy_kWh_per_kg * g_to_kg_conv + ely_stack_capex_EI # kg CO2e/kg H2
        electrolysis_Scope2_EI =  h2prod_grid_frac*scope2_grid_emissions_sum/h2prod_sum 
        electrolysis_Scope1_EI = 0
        electrolysis_total_EI  = electrolysis_Scope1_EI + electrolysis_Scope2_EI + electrolysis_Scope3_EI 
        electrolysis_total_EI_policy_grid = electrolysis_total_EI - wind_capex_EI * electrolyzer_energy_kWh_per_kg * g_to_kg_conv - ely_stack_capex_EI
        electrolysis_total_EI_policy_offgrid = 0
    elif grid_connection_scenario == 'grid-only':
        # Calculate grid-connected electrolysis emissions
        electrolysis_Scope3_EI = scope3_grid_emissions_sum/h2prod_sum  + ely_stack_capex_EI # kg CO2e/kg H2
        electrolysis_Scope2_EI = scope2_grid_emissions_sum/h2prod_sum 
        electrolysis_Scope1_EI = 0
        electrolysis_total_EI = electrolysis_Scope1_EI + electrolysis_Scope2_EI + electrolysis_Scope3_EI
        electrolysis_total_EI_policy_grid = electrolysis_total_EI - ely_stack_capex_EI
        electrolysis_total_EI_policy_offgrid = 0
    elif grid_connection_scenario == 'off-grid':    
        # Calculate renewable only electrolysis emissions        
        electrolysis_Scope3_EI = wind_capex_EI * electrolyzer_energy_kWh_per_kg * g_to_kg_conv + ely_stack_capex_EI # kg CO2e/kg H2
        electrolysis_Scope2_EI = 0
        electrolysis_Scope1_EI = 0
        electrolysis_total_EI = electrolysis_Scope1_EI + electrolysis_Scope2_EI + electrolysis_Scope3_EI
        electrolysis_total_EI_policy_offgrid = electrolysis_total_EI - wind_capex_EI * electrolyzer_energy_kWh_per_kg * g_to_kg_conv - ely_stack_capex_EI
        electrolysis_total_EI_policy_grid = 0
    
    return(electrolysis_total_EI_policy_grid,electrolysis_total_EI_policy_offgrid,h2prod_grid_frac)    


